/*
 * ------------------------------------------------------------------
 * NewPlane.cc - Plane tokens
 * Created by Robert Heller on Wed Oct 12 13:02:32 1994
 * ------------------------------------------------------------------
 * Modification History: $Log: NewPlane.cc,v $
// Revision 1.2  1994/10/12  17:03:12  heller
// Added RCS keyword hackery
//
 * ------------------------------------------------------------------
 * Contents:
 * ------------------------------------------------------------------
 *  
 * 
 *            Copyright 1994 University of Massachusetts.
 *                          All rights reserved.
 * 
 *      Permission to copy and modify this software and its documen-
 *      tation only for internal use in your organization is hereby
 *      granted, provided that this notice is retained thereon and
 *      on all copies.  UMASS makes no representations as to the sui-
 *      tability and operability of this software for any purpose.
 *      It is provided "as is" without express or implied warranty.
 * 
 *      UMASS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 *      INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
 *      NESS.  IN NO EVENT SHALL UMASS BE LIABLE FOR ANY SPECIAL,
 *      INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY OTHER DAMAGES WHAT-
 *      SOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 *      IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 *      ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PER-
 *      FORMANCE OF THIS SOFTWARE.
 * 
 *      No other rights, including, for example, the right to redis-
 *      tribute this software and its documentation or the right to
 *      prepare derivative works, are granted unless specifically
 *      provided in a separate license agreement.
 * 
 *      Copyright 1994, University of Massachusetts. All rights
 *      reserved.
 * 
 * 
 */

static char rcsid[] = "$Id";
#include <errno.h>
#include <NewPlane.h>
#include <llvs_plane.h>
#include <X11/Xutil.h>
/*
 * forward refs:  lower lever helper functions
 */

static void swap_longs(llvs_integer array[],int element_count);
static void swap_words(llvs_half_integer array[],int element_count);
static void cvt_floats(llvs_integer array[],int element_count,
		       int current_format,int desired_format);

/*
 * Generic Plane Code
 */

void Plane::FreeData()
{
  if (Data1D != NULL) {
    delete Data1D;
    Data1D = NULL;
  }
  if (Data2D != NULL) FreeData2D();
}

void Plane::FreeData2D()
{
  if (Data2D != NULL) {
    if (coord.YDir == Up) delete (char**) (Data2D-(coord.YOrigin));
    else delete (char**) (Data2D + (coord.YOrigin));

    Data2D = NULL;
  }
}

void Plane::SetData2D()
{
  if (Data2D != NULL) FreeData2D();

  Data2D = new unsigned char*[YSize];
  if (NULL == Data2D) {
    cerr << "Unable to allocate memory for Plane's Data2D field.\n";
    exit(-1);
  }

  for (int y = 0;y < YSize ; y++)
    {
      Data2D[y] = &Data1D[y*(XSize)*pixel_size]; // left edge of row
      if (coord.XDir == Up) Data2D[y] += (coord.XOrigin);		// XOrigin pixel of row
      else Data2D[y] -= (coord.XOrigin);
    }

  // adjust array base.
  if (coord.YDir == Up) Data2D += coord.YOrigin;
  else Data2D -= coord.YOrigin;
}

void Plane::InitData()
{
  // allocate image memory
  Data1D = new unsigned char[(XSize)*(YSize)*(pixel_size)];
  SetData2D();
}	

/*
 * Plane constructor
 * Mark plane_type with UNINITIALIZED
 */

Plane::Plane()
{
  coord.XOrigin = coord.YOrigin = XSize = YSize = Level = pixel_size = data_size = 0;
  coord.XDir = coord.YDir = Up;
  Data1D = NULL;
  Data2D = NULL;
  Name[0] = '\0';
  plane_type = UNINITIALIZED;
  ximage = NULL;
  minval = maxval = 0.0;
  for (int i=0;i < MAXGREY;i++) colorMapping[i] = 0xffffffff;
  DontScale = False;
  Incr_workId = None;
}

/*
 * Plane destructor.
 * free up allocated memory.
 */
		     
Plane::~Plane()
{
  FreeData();
  if (ximage != NULL) XDestroyImage(ximage);
  if (Incr_workId != None) XtRemoveWorkProc(Incr_workId);
  Incr_workId = None;
  Incr_refresh = False;
}


/*
 * base read plane function:  read in a plane and return a pointer to
 * a block of (converted) data.  Also return size, origin, and level.
 * returns NULL if plane type match error or if not a plane file.
 */

unsigned char* Plane::read_Data1D(Isristream& instream, int plane_type, 
			     int *xsize, int *ysize,
			     int *xloc, int *yloc, int *level,
			     AxisDirection *xdir, AxisDirection *ydir)
{
    /* plane file header record structure */
    static LLVS_PLANE_FILE_HEADER header;
    /* plane size header record structure */
    static LLVS_PLANE_SIZE_HEADER size_header;
    int plsize;			/* # bytes in plane */
    llvs_ubyte *data_pointer; /* pointer to data buffer */
    int rbytes, bytesleft;	/* I/O byte counters */
    int need_swap, need_cvt_float; /* flags to indicate if conversions needed */
    static scratchbuffer[BLOCKSIZE]; /* buffer */
    void* result;		/* result pointer */

    /* read header record */
    instream.read((char *)&header,sizeof(LLVS_PLANE_FILE_HEADER));
    if (!instream) return(NULL);
    /* conversion stuff */
    need_swap = header.bsex != LLVS_NATIVE_BYTE_SEX;
    need_cvt_float = header.floatfmt != LLVS_NATIVE_FLOATFMT;

    if (LLVS_XDIR_ISUP(header.axdir)) *xdir = Up;
    else *xdir = Down;

    if (LLVS_YDIR_ISUP(header.axdir)) *ydir = Up;
    else *ydir = Down;

    /* convert header to native format */
    if (need_swap) swap_longs(&header.pl_level,7);
    if (header.ptype == LLVS_PLF_FLOAT && need_cvt_float) 
	cvt_floats((long int*)&header.background,1,header.floatfmt,LLVS_NATIVE_FLOATFMT);
    /* gobble alist */
    bytesleft = header.alist_length; /* number of bytes to read */
    while (bytesleft > 0) {
	rbytes = bytesleft;
	if (rbytes > BLOCKSIZE) rbytes = BLOCKSIZE;
	instream.read((char *)scratchbuffer,rbytes);
	if (!instream) return(NULL);
	bytesleft -= rbytes;
	}
    /* read size header */
    instream.read((char *)&size_header,sizeof(LLVS_PLANE_SIZE_HEADER));
    if (!instream) return(NULL);
    /* convert header to native format */
    if (need_swap) swap_longs((long int*)&size_header,3);
    /* desired plane type? */
    if (header.ptype != plane_type)
    {
	// skip unread (bogus) bytes
	bytesleft = header.data_length - sizeof(LLVS_PLANE_SIZE_HEADER);
	while (bytesleft > 0) {
		rbytes = bytesleft;
		if (rbytes > BLOCKSIZE) rbytes = BLOCKSIZE;
		instream.read((char *)scratchbuffer,rbytes);
		if (!instream) return(NULL);
		bytesleft -= rbytes;
	}
    	errno = EINVAL;
    	return(NULL);
    }
    /* compute plane size */
    plsize = header.data_length - sizeof(LLVS_PLANE_SIZE_HEADER);
    result = (void*) new unsigned char[plsize];
    /* read in plane */
    data_pointer = (unsigned char *) result;
    bytesleft = plsize;
    while (bytesleft > 0) {
	rbytes = bytesleft;
	if (rbytes > BLOCKSIZE) rbytes = BLOCKSIZE;
	// read in a block
	instream.read((char *)data_pointer,rbytes);
	if (!instream)
	{
		int error = errno;
		delete result;
		errno = error;
		return(NULL);
	}
	// perform conversions as needed
	if (need_swap && header.ptype == LLVS_PLF_SHORT) 
	    swap_words((short int *)data_pointer,rbytes >> 1);
	if (need_swap && (header.ptype == LLVS_PLF_INT || 
			  header.ptype == LLVS_PLF_FLOAT))
	    swap_longs((long int *)data_pointer,rbytes >> 2);
	if (need_cvt_float && header.ptype == LLVS_PLF_FLOAT)
	    cvt_floats((long int *)data_pointer,rbytes >> 2,header.floatfmt,LLVS_NATIVE_FLOATFMT);
	data_pointer += rbytes;
	bytesleft -= rbytes;
	}
    // return size, origin, and level info
    *xsize = size_header.col_dimension;
    *ysize = size_header.row_dimension;
    *xloc  = header.col_location;
    *yloc  = header.row_location;
    *level = header.pl_level;
    // return pointer to plane data
    return((unsigned char*) result);
    }

/*
 * base write plane function.  Dump the data pointed to by datablock, with
 * all of the plane header glop.
 * Everything is written in *native* binary format.  The first few bytes of
 * the header specifies what the native format is (byte sex, floating point
 * format).
 */
    
int Plane::writebin(Isrostream& outstream)
{
    static LLVS_PLANE_FILE_HEADER header;
    /* plane size header record structure */
    static LLVS_PLANE_SIZE_HEADER size_header;
    unsigned char *data_pointer; /* pointer to data buffer */
    int rbytes, bytesleft;	/* I/O byte counters */

    /* fill in plane file header record */
    header.ptype = plane_type;
    header.bsex = LLVS_NATIVE_BYTE_SEX;
    header.floatfmt = LLVS_NATIVE_FLOATFMT;
    header.axdir = 0;	/* default to down (LLVS) */
    /* Set direction bits */
    if (coord.XDir == Up) SETF_LLVS_XDIR(header.axdir,LLVS_AX_DIR_UP);
    if (coord.YDir == Up) SETF_LLVS_YDIR(header.axdir,LLVS_AX_DIR_UP);
    header.pl_level = Level;
    header.row_location = coord.YOrigin;
    header.col_location = coord.XOrigin;
    header.background.iback = 0;
    header.data_length = data_size + sizeof(LLVS_PLANE_SIZE_HEADER);
    header.alist_length = 3; /* strlen("NIL"); */
    header.multi_plane_flag = 0;

    /* write header record */
    outstream.write((char *)&header,sizeof(LLVS_PLANE_FILE_HEADER));
    if (!outstream) return(-1);
    /* get pointer to associaions list */
    data_pointer = (llvs_ubyte *) "NIL";
    /* and length */
    bytesleft = header.alist_length;
    /* write out associations list */
    while (bytesleft > 0) {
	rbytes = bytesleft;
	if (rbytes > BLOCKSIZE) rbytes = BLOCKSIZE;
	outstream.write((char *)data_pointer,rbytes);
	if (!outstream) return(-1);
	data_pointer += rbytes;
	bytesleft -= rbytes;
	}
    /* fill in size header */
    size_header.datatype_again = header.ptype;
    size_header.row_dimension = YSize;
    size_header.col_dimension = XSize;

    /* write out size header */
    outstream.write((char *)&size_header,sizeof(LLVS_PLANE_SIZE_HEADER));
    if (!outstream) return(-1);
    /* get pointer & size of plane data */
    data_pointer = (unsigned char *) Data1D;
    bytesleft = data_size;
    /* write plane data */
    while (bytesleft > 0) {
	rbytes = bytesleft;
	if (rbytes > BLOCKSIZE) rbytes = BLOCKSIZE;
	outstream.write((char *)data_pointer,rbytes);
	if (!outstream) return(-1);
	data_pointer += rbytes;
	bytesleft -= rbytes;
	}
    return(0);
    }

/*
 * Plane binary reader.  First call generic plane reader function.
 * This function returns NULL on error conditions, including wrong plane
 * type.  Then adjust the plane size if necessary. Finally copy the data and
 * free the stratch buffer.
 */

void Plane::readbin(Isristream& stream)
{
  int xloc, yloc;
  AxisDirection xdir,ydir;

  FreeData();
  
  // generic read_plane()
  Data1D = read_Data1D(stream, plane_type, &XSize, &YSize,
		       &xloc, &yloc, &Level, &xdir,&ydir);
  if (NULL == Data1D) {
    cerr << "Unable to read plane.\n";
    exit(-1);
  }

  Data2D = new unsigned char*[YSize];
  if (NULL == Data2D) {
    cerr << "Unable to allocate memory for Data2D Plane pointer.\n";
    exit(-1);
  }

  if (coord.YDir == Up) Data2D += coord.YOrigin;
  else Data2D -= coord.YOrigin;
  data_size = XSize * YSize * pixel_size;

  // update level and location
  SetCoordinates(xloc,yloc,xdir,ydir);
}

/*
 * Plane formatted reader.  Read in the output of writefmt().
 * Snarf in header and check type and size.  Adjust size as needed.
 * Then snarf in pixels.
 */

void Plane::readfmt(Isristream& stream)
{
  int ix,iy;
  static char newname[18];
  char *p, ch;
  unsigned int newintpixel;
  float newfloatpixel;
  unsigned char *imdata_c;
  int *imdata_i;
  float *imdata_f;
  
  // snarf in header
  stream >> name();
  stream >> "[Level:";
  stream >> Level;
  stream >> ",XOrigin:";
  stream >> coord.XOrigin;
  stream >> ",YOrigin:";
  stream >> coord.YOrigin;
  stream >> ",XSize:";
  stream >> XSize;
  stream >> ",YSize:";
  stream >> YSize;
  stream >> ",XDir:";
  data_size = XSize * YSize * pixel_size;
  ch = stream.get();
  if (ch == 'U') {
    stream >> "p";
    coord.XDir = Up;
  } 
  else if (ch == 'D') {
    stream >> "own";
    coord.XDir = Down;
  }
  else {
    stream.putback(ch);
    stream >> "Up";
  }
  stream >> ",YDir:";
  ch = stream.get();
  if (ch == 'U') {
    stream >> "p";
    coord.YDir = Up;
  } 
  else if (ch == 'D') {
    stream >> "own";
    coord.YDir = Down;
  }
  else {
    stream.putback(ch);
    stream >> "Up";
  }
  stream >> ",Name:\"";
  for (p = newname;;p++) {
    ch = stream.get();
    if (ch == '"') {
      *p = '\0';
      break;
    } 
    else if (ch == '\\') {
      ch = stream.get();
      if (ch >= '0' && ch <= '7') {
	unsigned char ch1 = 0;
	do {
	  ch1 *= 8;
	  ch1 += ch - '0';
	  ch = stream.get();		
	} 
	while (ch >= '0' && ch <= '7');
	stream.putback(ch);
	ch = ch1;
      }
      *p++ = ch;
    } 
    else *p++ = ch;
  }
  stream >> "]:\n";
  // adjust size
  FreeData();
  InitData();
  SetCoordinates(coord.XOrigin,coord.YOrigin,coord.XDir,coord.YDir);
  
  /* snarf data (three cases -- one for each datatype) */
  if (plane_type == LLVS_PLF_BYTE) {
    for (imdata_c = (unsigned char *)Data1D, iy=0; iy < YSize;iy++) {
      stream >> newintpixel;
      for (ix=0;ix < XSize;ix++) {
	*imdata_c++ = (unsigned char) newintpixel;
	if (!stream) {
	  cerr << "error reading byte plane\n";
	  exit(-1);
	}
	if ((ix+1) < XSize) stream >> newintpixel;
	/* else stream >> "\n"; */
      }
    }
  }
  else if (plane_type == LLVS_PLF_INT) {
    for (imdata_i = (int *)Data1D, iy=0; iy < YSize;iy++) {
      stream >> newintpixel;
      for (ix=0;ix < XSize;ix++) {
	*imdata_i++ = newintpixel;
	if (!stream) {
	  cerr << "error reading byte plane\n";
	  exit(-1);
	}
	if ((ix+1) < XSize) stream >> newintpixel;
	/* else stream >> "\n"; */
      }
    }
  }
  else if (plane_type == LLVS_PLF_FLOAT) {
    for (imdata_f = (float *)Data1D, iy=0; iy < YSize;iy++) {
      stream >> newfloatpixel;
      for (ix=0;ix < XSize;ix++) {
	*imdata_f++ = newfloatpixel;
	if (!stream) {
	  cerr << "error reading byte plane\n";
	  exit(-1);
	}
	if ((ix+1) < XSize) stream >> newfloatpixel;
	/* else stream >> "\n"; */
      }
	  }
  }
  else {
    cerr << "Unimplemented Plane Type: " << plane_type << endl;
    exit(-1);
  }
}

void Plane::write_plane_header(Isrostream& stream)
{
  stream << name();
  stream << "[Level:" << Level << ",XOrigin:" << coord.XOrigin; 
  stream << ",YOrigin:" << coord.YOrigin << ",XSize:" << XSize << ",YSize:" << YSize;
  stream << ",XDir:";
  if (coord.XDir==Up) stream << "Up";
  else stream << "Down";
  stream << ",YDir:";
  if (coord.YDir==Up) stream << "Up";
  else stream << "Down";
  stream << ",Name:\"";	
  // header continued:  escape strange characters
  for (char* p=Name;*p != '\0';p++)
    {
      if (*p == '"' ||
	  *p == '\\')
	{
	  stream << '\\';
	  stream << *p;
	}
      else if (isprint(*p)) 
	stream << *p;
    }
  stream << (char*)"\"]:\n";
}  

void Plane::writefmt(Isrostream& stream)
{
  register unsigned char* imdata_c;
  register int* imdata_i;
  register float* imdata_f;
  register int iy,ix;
  write_plane_header(stream);

  if (plane_type == LLVS_PLF_BYTE) {
    for (imdata_c = (unsigned char *)Data1D,iy=0; iy < YSize;iy++)
      {
	for (ix=0; ix < XSize;ix++)
	  {
	    stream << ' ';
	    stream << (short) *imdata_c++;
	  }
	stream << '\n';
      }
  }
  else if (plane_type == LLVS_PLF_INT) {
    for (imdata_i = (int *)Data1D,iy=0; iy < YSize;iy++)
      {
	for (ix=0; ix < XSize;ix++)
	  {
	    stream << ' ';
	    stream << *imdata_i++;
	  }
	stream << '\n';
      }
  }
  else if (plane_type == LLVS_PLF_FLOAT) {
    for (imdata_f = (float *)Data1D,iy=0; iy < YSize;iy++)
      {
	for (ix=0; ix < XSize;ix++)
	  {
	    stream << ' ';
	    stream << *imdata_f++;
	  }
	stream << '\n';
      }
  }
  else {
    cerr << "Writing unknown plane type: " << plane_type << endl;
    exit(-1);
  }
}

void Plane::SetCoordinates(int xorigin,int yorigin,
			   AxisDirection xdir,AxisDirection ydir)
{
	int y;

	// reset Data2D pointer to its real base
	if (coord.YDir == Up) Data2D -= coord.YOrigin;
	else Data2D += coord.YOrigin;

	// set new coords
	coord.XOrigin = xorigin;
	coord.YOrigin = yorigin;
	coord.XDir    = xdir;
	coord.YDir    = ydir;

	// initialize the real rows
	for (y = 0;y < YSize ; y++)
	{
		Data2D[y] = &Data1D[y*(XSize)*pixel_size];   // left edge of row
		if (coord.XDir == Up) Data2D[y] += (coord.XOrigin); // XOrigin pixel of row
		else Data2D[y] -= (coord.XOrigin);
	}
	// adjust array base.
	if (coord.YDir == Up) Data2D += coord.YOrigin;
	else Data2D -= coord.YOrigin;
}

void Plane::GetMinMax(int &minx,int &maxx,int &miny,int &maxy)
{
	int x1,x2,y1,y2;

	x1 = coord.XOrigin;
	if (coord.XDir == Up) x2 = x1 - (XSize-1);
	else x2 = x1 + (XSize-1);
	if (x1 < x2)
	{
		minx = x1;
		maxx = x2;
	} else
	{
		minx = x2;
		maxx = x1;
	}

	y1 = coord.YOrigin;
	if (coord.YDir == Up) y2 = y1 - (YSize-1);
	else y2 = y1 + (YSize-1);
	if (y1 < y2)
	{
		miny = y1;
		maxy = y2;
	} else
	{
		miny = y2;
		maxy = y1;
	}

}

void Plane::GetMin(int &minx,int &miny)
{
	int x1,x2,y1,y2;

	x1 = coord.XOrigin;
	if (coord.XDir == Up) x2 = x1 - XSize;
	else x2 = x1 + XSize;
	if (x1 < x2) minx = x1;
	else minx = x2;

	y1 = coord.YOrigin;
	if (coord.YDir == Up) y2 = y1 - YSize;
	else y2 = y1 + YSize;
	if (y1 < y2) miny = y1;
	else miny = y2;
}

void Plane::GetMax(int &maxx,int &maxy)
{
	int x1,x2,y1,y2;

	x1 = coord.XOrigin;
	if (coord.XDir == Up) x2 = x1 - XSize;
	else x2 = x1 + XSize;
	if (x1 < x2) maxx = x2;
	else maxx = x1;

	y1 = coord.YOrigin;
	if (coord.YDir == Up) y2 = y1 -  YSize;
	else y2 = y1 + YSize;
	if (y1 < y2) maxy = y2;
	else maxy = y1;
}

/*
 * generic input function.  Branch on stream type
 */

void Plane::input(Isristream& stream)
{
  /* if plane is unitialized, initialize it. Note that this
     would be better done by explicitly calling the subclasses'
     constructors, btu I haven't figured out how to do that yet. */
  if (plane_type == UNINITIALIZED) {
    char ch = stream.get();
    stream.putback(ch);
    switch ((int) ch) 
      {
      case LLVS_PLF_BYTE  : plane_type = LLVS_PLF_BYTE;
			    pixel_size = sizeof(unsigned char);
			    Init(0, 0, 0, 0, Up, Up, "");
			    break;
      case LLVS_PLF_INT   : plane_type = LLVS_PLF_INT;
			    pixel_size = sizeof(int);
			    Init(0, 0, 0, 0, Up, Up, "");
			    break;
      case LLVS_PLF_FLOAT : plane_type = LLVS_PLF_FLOAT;
			    pixel_size = sizeof(float);
			    Init(0, 0, 0, 0, Up, Up, "");
			    break;
      case LLVS_PLF_BIT   : cout << "Bit Planes not implemented\n";
                       exit(-1);
      case LLVS_PLF_SHORT : cout << "Short Planes not implemented\n";
      default : cout << "unknown plane type\n";
                exit(-1);
      }
  }
  
  if (stream.format == BINARY) readbin(stream);
  else readfmt(stream);
}

/*
 * generic output function.  Branch on stream type
 */

void Plane::plane_output(Isrostream& stream)
{
  if (stream.format == BINARY) writebin(stream);
  else writefmt(stream);
}

void Plane::output(Isrostream& stream)
{
  plane_output(stream);
}


int calculate_image_level(int XSize, int YSize)
{
  int MinSize, level = 0;

  if (XSize > YSize) MinSize = YSize;
  else MinSize = XSize;

  while (MinSize > pow(2.0, level)) level++;
  return(level);
}

/* once plane_type and pixel_size have been filled in, initialize
   the rest of a plane structure. */

void Plane::Init (int xsize, int ysize, int xorigin, int yorigin, 
		  AxisDirection xdir, AxisDirection ydir, char *str)
{
  // fill in the various fields
  if (str != NULL) {
    strncpy(Name,str,sizeof(Name));
    Name[sizeof(Name)-1] = '\0';
  }
  else Name[0] = '\0';

  coord.XOrigin   = xorigin;
  coord.YOrigin   = yorigin;
  XSize     = xsize;
  YSize     = ysize;
  Level     = calculate_image_level(XSize, YSize);
  coord.XDir = xdir;
  coord.YDir = ydir;
  data_size = pixel_size * xsize * ysize;
  InitData();
}

/*
 * BytePlane constructor.
 *
 * Create and initialize a BytePlane.  Memory is allocated for the pixels
 * (Data1D slot) and for the row pointers (Data2D slot).  The pixel data
 * array is wide enough for the image width (XSize) plus twice the boundry
 * size (BSize) and as tall as the image (YSize).
 * The row pointer array is as tall as the image (YSize) plus twice the
 * boundry size (BSize).
 * The row pointer array is adjusted so the 0th element is actualy the
 * (YOrigin+BSize)'th element, and its pointer values point to the
 * (XOrigin+BSize)'th element of the row.  Most of the offset math is
 * effectively pre-computed here.  Only the reverse directionality of the
 * coordinates are handled at access time (see the inline pixel function).
 */

BytePlane::BytePlane ()
{
  plane_type = LLVS_PLF_BYTE;
  pixel_size = sizeof(unsigned char);
  Init(0, 0, 0, 0, Up, Up, "");
}

BytePlane::BytePlane (int xsize, int ysize, int xorigin, int yorigin, 
		      AxisDirection xdir/* = Up*/, AxisDirection ydir/* = Up*/, char *str/* = ""*/)
{
  // fill in the various fields
  plane_type = LLVS_PLF_BYTE;
  pixel_size = sizeof(unsigned char);
  Init(xsize, ysize, xorigin, yorigin, xdir, ydir, str);
}

BytePlane::BytePlane(Plane &pl, char *str/* = ""*/)
{
	plane_type = LLVS_PLF_BYTE;
	pixel_size = sizeof(unsigned char);
	Init(pl.XSize, pl.YSize, pl.coord.XOrigin, pl.coord.YOrigin, 
	     pl.coord.XDir, pl.coord.YDir, str);
}

BytePlane::~BytePlane()
{
  FreeData();
}

/*
 * generic input function.  Branch on stream type
 */

void BytePlane::input(Isristream& stream)
{
  if (stream.format == BINARY) readbin(stream);
  else readfmt(stream);
}

/*
 * generic output function.  Branch on stream type
 */

void BytePlane::output(Isrostream& stream)
{
	plane_output(stream);
}

/* 
 * overload >>, since planes are a non-standard format
 */

BytePlane *read_byte_plane(Isristream& is)
{
  BytePlane *data;

  data = new BytePlane;
  data->input(is);
  return data;
}



/*
 * IntPlane constructor.  Just like BytePlane constructor.
 */

IntPlane::IntPlane()
{
  plane_type = LLVS_PLF_INT;
  pixel_size = sizeof(int);
  Init(0, 0, 0, 0, Up, Up, "");
}

IntPlane::IntPlane (int xsize, int ysize, int xorigin, int yorigin, 
		      AxisDirection xdir/* = Up*/, AxisDirection ydir/* = Up*/, char *str/* = ""*/)
{
  // fill in the various fields
  plane_type = LLVS_PLF_INT;
  pixel_size = sizeof(int);
  Init(xsize, ysize, xorigin, yorigin, xdir, ydir, str);
}

IntPlane::IntPlane(Plane &pl, char *str/* = ""*/)
{
	plane_type = LLVS_PLF_INT;
	pixel_size = sizeof(int);
	Init(pl.XSize, pl.YSize, pl.coord.XOrigin, pl.coord.YOrigin, 
	     pl.coord.XDir, pl.coord.YDir, str);
}

IntPlane::~IntPlane()
{
  FreeData();
}

/*
 * generic input function.  Branch on stream type
 */

void IntPlane::input(Isristream& stream)
{
  if (stream.format == BINARY) readbin(stream);
  else readfmt(stream);
}

/*
 * generic output function.  Branch on stream type
 */

void IntPlane::output(Isrostream& stream)
{
  plane_output(stream);
}
IntPlane *read_int_plane(Isristream& is)
{
  IntPlane *data;

  data = new IntPlane;
  data->input(is);
  return data;
}

/*
 * FloatPlane constructor.  Just like BytePlane constructor.
 */

FloatPlane::FloatPlane()
{
  plane_type = LLVS_PLF_FLOAT;
  pixel_size = sizeof(float);
  Init(0, 0, 0, 0, Up, Up, "");
}

FloatPlane::FloatPlane (int xsize, int ysize, int xorigin, int yorigin, 
		      AxisDirection xdir/* = Up*/, AxisDirection ydir/* = Up*/, char *str/* = ""*/)
{
  // fill in the various fields
  plane_type = LLVS_PLF_FLOAT;
  pixel_size = sizeof(float);
  Init(xsize, ysize, xorigin, yorigin, xdir, ydir, str);
}

FloatPlane::FloatPlane(Plane &pl, char *str/* = ""*/)
{
	plane_type = LLVS_PLF_FLOAT;
	pixel_size = sizeof(float);
	Init(pl.XSize, pl.YSize, pl.coord.XOrigin, pl.coord.YOrigin, 
	     pl.coord.XDir, pl.coord.YDir, str);
}

FloatPlane::~FloatPlane()
{
  FreeData();
}

/*
 * generic input function.  Branch on stream type
 */

void FloatPlane::input(Isristream& stream)
{
  if (stream.format == BINARY) readbin(stream);
  else readfmt(stream);
}

/*
 * generic output function.  Branch on stream type
 */

void FloatPlane::output(Isrostream& stream)
{
  plane_output(stream);
}

FloatPlane *read_float_plane(Isristream& is)
{
  FloatPlane *data;

  data = new FloatPlane;
  data->input(is);
  return data;
}

#ifdef GETBIT
#undef GETBIT
#endif
#define GETBIT(base,offset) \
((base)[(offset) >> 3] >>((offset) & 0x07) & 0x01)

BytePlane *read_bit_plane_as_byte_plane(Isristream& is)
{
  int xloc, yloc;
  AxisDirection xdir,ydir;
  int xsize,ysize,level;

  BytePlane *data;
  data = new BytePlane;
  unsigned char *bits = data->read_Data1D(is, LLVS_PLF_BIT, &xsize,
				          &ysize, &xloc, &yloc, &level,
				          &xdir, &ydir);
  if (NULL == bits) {
    cerr << "Unable to read plane.\n";
    exit(-1);
  }
  data->Init(xsize, ysize, xloc, yloc, xdir, ydir, NULL);

  unsigned char *rowptr = data->Data1D;
  int roff = 0;
  for (int iy = 0; iy < ysize; iy++)
  {
    unsigned char *colptr = rowptr;
    int coff = roff;
    for (int ix = 0; ix < xsize; ix++)
    {
      *colptr++ = GETBIT(bits,coff);
      coff++;
    }
    rowptr += xsize;
    roff   += xsize;
  }
  delete bits;
  return data;
}

IntPlane *read_short_plane_as_int_plane(Isristream& is)
{
  int xloc, yloc;
  AxisDirection xdir,ydir;
  int xsize,ysize,level;

  IntPlane *data;
  data = new IntPlane;
  short int  *words = (short int*) data->read_Data1D(is, LLVS_PLF_SHORT, &xsize,
				   		     &ysize, &xloc, &yloc, 
						     &level, &xdir, &ydir);
  if (NULL == words) {
    cerr << "Unable to read plane.\n";
    exit(-1);
  }
  data->Init(xsize, ysize, xloc, yloc, xdir, ydir, NULL);

  long int  *rowptr = (long int*) data->Data1D;
  int roff = 0;
  for (int iy = 0; iy < ysize; iy++)
  {
    long int  *colptr = rowptr;
    int coff = roff;
    for (int ix = 0; ix < xsize; ix++)
    {
      *colptr++ = *(words + coff);
      coff++;
    }
    rowptr += xsize;
    roff   += xsize;
  }
  delete words;
  return data;
}


Token *read_llvs_plane(Isristream& is)
{
  char ch = is.get();
  is.putback(ch);
  switch ((int) ch) 
    {
      case LLVS_PLF_BYTE  : return read_byte_plane(is);
      case LLVS_PLF_INT   : return read_int_plane(is);
      case LLVS_PLF_FLOAT : return read_float_plane(is);
      case LLVS_PLF_BIT   : return read_bit_plane_as_byte_plane(is);
      case LLVS_PLF_SHORT : return read_short_plane_as_int_plane(is);
      default : cout << "unknown plane type\n";
                exit(-1);
    }
  return NULL;
}

void copy_plane_header(Plane *source, Plane *dest)
{
  (*dest) = *source;
}

void Plane::BitDraw(Display* display, Drawable drawable, GC gc,
			Transform2D* trans, float xmin, float ymin, 
			float xmax, float ymax, Visual *visual,
		    unsigned int depth,Colormap cmap)
{
	Boolean DoIncr_PutImParams = False;
	double *wsize = trans->Apply((double)XSize,(double)YSize),
	       *wzero = trans->Apply(0.0,0.0);
	double winwidth = fabs(wsize[0]-wzero[0]),
	       winheight = fabs(wsize[1]-wzero[1]);
	int XIMWidth = (int) winwidth,
	    XIMHeight = (int) winheight;
	/*if (XIMWidth < winwidth) XIMWidth++;*/
	/*if (XIMHeight < winheight) XIMHeight++;*/
/*	cerr << "*** Plane::BitDraw(): winwidth = " << winwidth << ", winheight = " << winheight << "\n";*/
/*	cerr << "***                -: XIMWidth = " << XIMWidth << ", XIMHeight = " << XIMHeight << "\n";*/
	delete wsize;
	delete wzero;
	if (ximage != NULL)
	{
		/*cerr << "***                -: ximage->width = " << ximage->width << ", ximage->height = " << ximage->height << "\n";*/
		if (ximage->width != XIMWidth || ximage->height != XIMHeight)
		{
			XDestroyImage(ximage);
			ximage = NULL;
			if (Incr_workId != None)
			{
				XtRemoveWorkProc(Incr_workId);
				Incr_workId = None;
				Incr_refresh = False;
			}
		}
	}
	if (ximage == NULL)
	{
		char *data;
		if (depth == 1)
		{
			int a = XIMWidth+7;
			int b = a / 8;
			int d = XIMHeight*b;
			data = (char*) calloc(d,sizeof(unsigned char));
			if (data == NULL)
			{
				int err = errno;
				cerr << "Out of memory in NewPlane::BitDraw\n";
				exit(err);
			}
		} else if (depth == 8)
		{
			int a = XIMHeight*XIMWidth;
			data = (char*) calloc(a,sizeof(unsigned char));
			if (data == NULL)
			{
				int err = errno;
				cerr << "Out of memory in NewPlane::BitDraw\n";
				exit(err);
			}
		} else if (depth == 24)
		{
			int a = XIMHeight*XIMWidth*3;
			data = (char*) calloc(a,sizeof(unsigned char));
			if (data == NULL)
			{
				int err = errno;
				cerr << "Out of memory in NewPlane::BitDraw\n";
				exit(err);
			}
		}
		ximage = XCreateImage(display,visual,depth,ZPixmap,0,data,
				      XIMWidth,XIMHeight,8,0);
		if (ximage == NULL)
		{
			int err = errno;
			cerr << "Out of memory in NewPlane::BitDraw\n";
			exit(err);
		}
		if (XIMHeight > XIMThresh || XIMWidth > XIMThresh)
		{
			/* cerr << "*** Plane::BitDraw(): Incremental draw\n"; */
			DoIncr_PutImParams = True;
			if (depth == 1) Incr_ditherBitImage(display,XIMWidth,XIMHeight,cmap,trans);
			else Incr_drawGreyImage(display,XIMWidth,XIMHeight,cmap,trans);
		} else
		{
			Incr_refresh = False;
			/*cerr << "*** Plane::BitDraw(): Not Incremental draw\n";*/
			Incr_workId = None;
			if (depth == 1) ditherBitImage(display,XIMWidth,XIMHeight,cmap,trans);
			else drawGreyImage(display,XIMWidth,XIMHeight,cmap,trans);
		}
	}
	double *c1 = trans->Apply(xmin,ymin),
	       *c2 = trans->Apply(xmax,ymax);
	int imw = (int)fabs(c1[0]-c2[0]),
	    imh = (int)fabs(c1[1]-c2[1]);
	int hpix = (int) ((XIMWidth / (double) XSize) + 0.5),
	    vpix = (int) ((XIMHeight / (double) YSize) + 0.5);
	int isrcx = (int) ((xmin - coord.XOrigin) * hpix),
	    isrcy = (int) ((ymin - coord.YOrigin) * vpix);
	if (coord.XDir == Up)
	{
		isrcx = (int) (((-xmin) + coord.XOrigin) * hpix);
	}
	if (coord.YDir == Up)
	{
		isrcy = (int) (((-ymin) + coord.YOrigin) * vpix);
	}		
	if ((imw+isrcx) > XIMWidth) imw = XIMWidth-isrcx;
	if ((imh+isrcy) > XIMHeight) imh = XIMHeight-isrcy;
	int idestx = (int) c1[0],
	    idesty = (int) c1[1];
	delete c1;
	delete c2;
	if (imw <= 0 || imh <= 0) return;
/*	cerr << "***                -: c1[0] = " << c1[0] << ", c1[1] = " << c1[1] << "\n";*/
/*	cerr << "***                -: c2[0] = " << c2[0] << ", c2[1] = " << c2[1] << "\n";*/
/*	cerr << "***                -: imw = " << imw << ", imh = " << imh << "\n";*/
/*	cerr << "***                -: isrcx = " << isrcx << ", isrcy = " << isrcy << "\n";*/
	if (DoIncr_PutImParams)
	{
		Incr_PutImParams.isrcx = isrcx;
		Incr_PutImParams.isrcy = isrcy;
		Incr_PutImParams.idestx = idestx;
		Incr_PutImParams.idesty = idesty;
		Incr_PutImParams.imw = imw;
		Incr_PutImParams.imh = imh;
		Incr_drawable = drawable;
		Incr_gc = gc;
		Incr_refresh = True;
	} else
	{
		XPutImage(display,drawable,gc,ximage,
		  isrcx,isrcy, /* src */
		  idestx,idesty, /* dest */
		  imw, imh		/* size */
		  );
		if (Incr_workId)
		{
			if ((idestx <= Incr_PutImParams.idestx &&
			     (idestx+imw) >= 
			     (Incr_PutImParams.idestx + Incr_PutImParams.imw)) &&
			    (idesty <= Incr_PutImParams.idesty &&
			     (idesty+imh) >= 
			     (Incr_PutImParams.idesty + Incr_PutImParams.imh)))
			{
				Incr_PutImParams.isrcx = isrcx;
				Incr_PutImParams.isrcy = isrcy;
				Incr_PutImParams.idestx = idestx;
				Incr_PutImParams.idesty = idesty;
				Incr_PutImParams.imw = imw;
				Incr_PutImParams.imh = imh;
				Incr_drawable = drawable;
				Incr_gc = gc;
				Incr_refresh = True;
			}
		}
	}
			
				
}

void  Plane::plane_min_max()
{
	int flag;
	long int ipixels;

	flag = TRUE;
	ipixels = XSize * YSize;
	switch (plane_type)
	{
		case LLVS_PLF_BYTE:
		for (unsigned char *bptr = (unsigned char *)Data1D;ipixels > 0;bptr++,ipixels--)
		{
			if (flag)
			{
				minval = (double) *bptr;
				maxval = minval;
				flag = FALSE;
			} else if (((double) *bptr) > maxval) maxval = (double) *bptr;
			else if (((double) *bptr) < minval) minval = (double) *bptr;
		}
		break;
		case LLVS_PLF_INT:
		for (long int *lptr = (long int *)Data1D;ipixels > 0;lptr++,ipixels--)
		{
			if (flag)
			{
				minval = (double) *lptr;
				maxval = minval;
				flag = FALSE;
			} else if (((double) *lptr) > maxval) maxval = (double) *lptr;
			else if (((double) *lptr) < minval) minval = (double) *lptr;
		}
		break;
		case LLVS_PLF_FLOAT:
		for (float *fptr = (float *)Data1D;ipixels > 0;fptr++,ipixels--)
		{
			if (flag)
			{
				minval = (double) *fptr;
				maxval = minval;
				flag = FALSE;
			} else if (((double) *fptr) > maxval) maxval = (double) *fptr;
			else if (((double) *fptr) < minval) minval = (double) *fptr;
		}
		break;
	}
}


/* maximum device intensity.  these values are typical for 7 bits/color
   devices */

#define ZERO 0
#define MAXINT_F 255.0
#define MAXINT_I 255

void  Plane::drawGreyImage(Display *d,int XIMWidth,int XIMHeight,Colormap cmap,Transform2D* trans)
{
	int hpixsize,vpixsize;
	int xim_y_coord, xim_x_coord;
	int crow,ccol;
	int byte, ipixrow, ipixcol;
	double intensity_scale,value_range,pixel,scaled_pixel;
	unsigned char *row_base;
	XColor xcolor;

	if (minval == 0.0 && maxval == 0.0) plane_min_max();
	
	hpixsize = (int) ((XIMWidth / (double) XSize) + 0.5);
	vpixsize = (int) ((XIMHeight / (double) YSize) + 0.5);
	value_range = maxval - minval;
	intensity_scale = MAXINT_F / value_range;
	for (crow = 0; crow < YSize; crow++)
	{
		row_base = (crow * XSize * pixel_size) + Data1D;
		xim_y_coord = crow * vpixsize;
		
		for (ccol = 0; ccol < XSize; ccol++)
		{
			xim_x_coord = ccol * hpixsize;
			switch (plane_type)
			{
				case LLVS_PLF_BYTE:
					pixel = (double) *(row_base + ccol);
					break;
				case LLVS_PLF_INT:
					pixel = (double) *(((long int *)row_base) + ccol);
					break;
				case LLVS_PLF_FLOAT:
					pixel = (double) *(((float *)row_base) + ccol);
					break;
			}
			if (DontScale) byte = (int) pixel;
			else {
			  /* scale pixel.  out of range values get black or white.
			     otherwise linear mapping */
			  if (intensity_scale > 0.0) {
			    /* positive slope case */
			    if (pixel < minval)
				scaled_pixel = 0.0;
			    else if (pixel > maxval)
				scaled_pixel = MAXINT_F;
			    else
				scaled_pixel =(pixel - minval) *
				    intensity_scale;
			  }
			  else {	/* negative slope case */
			    if (pixel < maxval)
				scaled_pixel = MAXINT_F;
			    else if (pixel > minval)
				scaled_pixel = 0.0;
			    else
				scaled_pixel =(pixel - minval) *
				    intensity_scale;
			  }
			  byte = ((int)(scaled_pixel + 0.5)) >> 1;
			}
			if (colorMapping[byte] == 0xffffffff)
			{
				xcolor.red = byte << 8;
				xcolor.green = xcolor.red;
				xcolor.blue = xcolor.red;
				xcolor.flags = DoRed|DoGreen|DoBlue;
				XAllocColor(d,cmap,&xcolor);
				colorMapping[byte] = xcolor.pixel;
			}
			for (ipixrow=0;ipixrow<vpixsize;ipixrow++) {
				for (ipixcol=0;ipixcol<hpixsize;ipixcol++) {
					XPutPixel(ximage,xim_x_coord+ipixcol,xim_y_coord+ipixrow,colorMapping[byte]);
				}
			}
		}
	}
}



 /* dithering matrix */
static const short int matrix[16][16] = {
    {1, 235, 59, 219, 15, 231, 55, 215, 4, 234, 58, 218, 14, 230, 54, 214},
    {129, 65, 187, 123, 143, 79, 183, 119, 132, 68, 186, 122, 142, 78, 182,
     118},
    {33, 193, 17, 251, 47, 207, 31, 247, 36, 196, 20, 250, 46, 206, 30, 246},
    {161, 97, 145, 81, 175, 111, 159, 95, 164, 100, 148, 84, 174, 110, 158,
     94},
    {9, 225, 49, 209, 5, 239, 63, 223, 12, 228, 52, 212, 8, 238, 62, 222},
    {137, 73, 177, 113, 133, 69, 191, 127, 140, 76, 180, 116, 136, 72, 190,
     126},
    {41, 201, 25, 241, 37, 197, 21, 255, 44, 204, 28, 244, 40, 200, 24, 254},
    {169, 105, 153, 89, 165, 101, 149, 85, 172, 108, 156, 92, 168, 104, 152,
     88},
    {3, 233, 57, 217, 13, 229, 53, 213, 2, 236, 60, 220, 16, 232, 56, 216},
    {131, 67, 185, 121, 141, 77, 181, 117, 130, 66, 188, 124, 144, 80, 184,
     120},
    {35, 195, 19, 249, 45, 205, 29, 245, 34, 194, 18, 252, 48, 208, 32, 248},
    {163, 99, 147, 83, 173, 109, 157, 93, 162, 98, 146, 82, 176, 112, 160,
     96},
    {11, 227, 51, 211, 7, 237, 61, 221, 10, 226, 50, 210, 6, 240, 64, 224},
    {139, 75, 179, 115, 135, 71, 189, 125, 138, 74, 178, 114, 134, 70, 192,
     128},
    {43, 203, 27, 243, 39, 199, 23, 253, 42, 202, 26, 242, 38, 198, 22, 256},
    {171, 107, 155, 91, 167, 103, 151, 87, 170, 106, 154, 90, 166, 102, 150,
     86}
    };




void  Plane::ditherBitImage(Display *d,int XIMWidth,int XIMHeight,Colormap cmap,Transform2D* trans)
{
	int xim_y_coord, xim_x_coord, bitoff, bitbyte;
	register int maskrow, maskcol, saved_maskcol, saved_maskrow;
	register int ipixrow, ipixcol;
	int crow,ccol;
	int byte;
	double intensity_scale,value_range,pixel,scaled_pixel;
	register short int *maskrow_ptr;
	int hpixsize,vpixsize;
	unsigned char *row_base;
/*
	Pixel Black, White;
	XColor defcolor,exactcolor;
	XAllocNamedColor(d,cmap,"Black",&defcolor,&exactcolor);
	Black = exactcolor.pixel;
	XAllocNamedColor(d,cmap,"White",&defcolor,&exactcolor);
	White = exactcolor.pixel;
*/

	if (minval == 0.0 && maxval == 0.0) plane_min_max();

	hpixsize = (int) ((XIMWidth / (double) XSize) + 0.5);
	vpixsize = (int) ((XIMHeight / (double) YSize) + 0.5);
	saved_maskrow = maskrow = 0;
	value_range = maxval - minval;
	intensity_scale = MAXINT_F / value_range;
	
	for (crow = 0; crow < YSize; crow++)
	{
		row_base = (crow * XSize * pixel_size) + Data1D;
		xim_y_coord = crow * vpixsize;
		saved_maskrow = (maskrow % 16);
		maskcol = 0;
		
		for (ccol = 0; ccol < XSize; ccol++)
		{
			xim_x_coord = ccol * hpixsize;
			switch (plane_type)
			{
				case LLVS_PLF_BYTE:
					pixel = (double) *(row_base + ccol);
					break;
				case LLVS_PLF_INT:
					pixel = (double) *(((long int *)row_base) + ccol);
					break;
				case LLVS_PLF_FLOAT:
					pixel = (double) *(((float *)row_base) + ccol);
					break;
			}
			/* scale pixel.  out of range values get black or white.
			   otherwise linear mapping */
			if (intensity_scale > 0.0) {
			    /* positive slope case */
			    if (pixel < minval)
				scaled_pixel = 0.0;
			    else if (pixel > maxval)
				scaled_pixel = MAXINT_F;
			    else
				scaled_pixel =(pixel - minval) *
				    intensity_scale;
			}
			else {	/* negative slope case */
			    if (pixel < maxval)
				scaled_pixel = MAXINT_F;
			    else if (pixel > minval)
				scaled_pixel = 0.0;
			    else
				scaled_pixel =(pixel - minval) *
				    intensity_scale;
			}
			byte = (int)(scaled_pixel + 0.5);
			saved_maskcol = (maskcol % 16);
			for (ipixrow=0;ipixrow<vpixsize;ipixrow++) {
				bitbyte = 1 << (bitoff & 0x07);
				if (maskrow > 15) maskrow = 0;
				maskrow_ptr = (short int *)(&matrix[maskrow][0]);
				maskcol = saved_maskcol;
				for (ipixcol=0;ipixcol<hpixsize;ipixcol++) {
					if (maskcol > 15) maskcol = 0;
					if (maskrow_ptr[maskcol] < byte) {
						XPutPixel(ximage,xim_x_coord+ipixcol,xim_y_coord+ipixrow,1);
					}/* else	XPutPixel(ximage,xim_x_coord+ipixcol,xim_y_coord+ipixrow,White);*/
					bitbyte = bitbyte << 1;
					if (bitbyte > 0x0080) {
						bitbyte = 1;
					}
					maskcol++;
				}
				maskrow++;
			}
		}
	}
}

extern XtAppContext app_con;	/* from xisrdisplay's main.cc OR DummyAppCont.cc */

void  Plane::Incr_drawGreyImage(Display *d,int XIMWidth,int XIMHeight,Colormap cmap,Transform2D* trans)
{
	double value_range;

	if (minval == 0.0 && maxval == 0.0) plane_min_max();

	Incr_hpixsize = (int) ((XIMWidth / (double) XSize) + 0.5);
	Incr_vpixsize = (int) ((XIMHeight / (double) YSize) + 0.5);
	value_range = maxval - minval;
	Incr_intensity_scale = MAXINT_F / value_range;
	Incr_Currow = 0;
	Incr_cmap = cmap;
	Incr_display = d;
	Incr_dithering = False;
	Incr_workId = XtAppAddWorkProc(app_con,
				Incr_PlaneBitDrawWorkCallback,(XtPointer) this);
	/* cerr << "*** Plane::Incr_drawGreyImage() Incr_workId = "
	     << Incr_workId << "\n";*/
}

void  Plane::Incr_ditherBitImage(Display *d,int XIMWidth,int XIMHeight,Colormap cmap,Transform2D* trans)
{
	double value_range;

	if (minval == 0.0 && maxval == 0.0) plane_min_max();

	Incr_hpixsize = (int) ((XIMWidth / (double) XSize) + 0.5);
	Incr_vpixsize = (int) ((XIMHeight / (double) YSize) + 0.5);
	value_range = maxval - minval;
	Incr_intensity_scale = MAXINT_F / value_range;
	Incr_display = d;
	Incr_Currow = 0;
	Incr_maskrow = 0;
	Incr_dithering = True;
	Incr_workId = XtAppAddWorkProc(app_con,
				Incr_PlaneBitDrawWorkCallback,(XtPointer) this);
	/* cerr << "*** Plane::Incr_ditherBitImage() Incr_workId = "
	     << Incr_workId << "\n"; */
}

Boolean Incr_PlaneBitDrawWorkCallback(XtPointer clientData)
{
	Plane *plane = (Plane*) clientData;
	Boolean workDone;
	if (plane->Incr_dithering) workDone = plane->Incr_ditherBitImageCallback();
	else workDone = plane->Incr_drawGreyImageCallback();
	if (workDone) plane->Incr_workId = None;
	return(workDone);
}

Boolean Plane::Incr_drawGreyImageCallback()
{
	int hpixsize,vpixsize;
	int xim_y_coord, xim_x_coord;
	int crow,ccol,numrows,stoprow,numcols,stopcol;
	int byte, ipixrow, ipixcol;
	double intensity_scale,pixel,scaled_pixel;
	unsigned char *row_base;
	XColor xcolor;
	Colormap cmap;
	Display *d;

	hpixsize = Incr_hpixsize;
	vpixsize = Incr_vpixsize;
	intensity_scale = Incr_intensity_scale;
	cmap = Incr_cmap;
	d = Incr_display;
	numrows = XIMThresh / vpixsize;
	numcols = XIMThresh / hpixsize;
	stoprow = Incr_Currow + numrows;
	stopcol = Incr_Curcol + numcols;
	if (stoprow > YSize) stoprow = YSize;
	if (stopcol > XSize) stopcol = XSize;

	for (crow = Incr_Currow; crow < stoprow; crow++)
	{
		row_base = (crow * XSize * pixel_size) + Data1D;
		xim_y_coord = crow * vpixsize;
		
		for (ccol = Incr_Curcol; ccol < stopcol; ccol++)
		{
			xim_x_coord = ccol * hpixsize;
			switch (plane_type)
			{
				case LLVS_PLF_BYTE:
					pixel = (double) *(row_base + ccol);
					break;
				case LLVS_PLF_INT:
					pixel = (double) *(((long int *)row_base) + ccol);
					break;
				case LLVS_PLF_FLOAT:
					pixel = (double) *(((float *)row_base) + ccol);
					break;
			}
			if (DontScale) byte = (int) pixel;
			else {
			  /* scale pixel.  out of range values get black or white.
			     otherwise linear mapping */
			  if (intensity_scale > 0.0) {
			    /* positive slope case */
			    if (pixel < minval)
				scaled_pixel = 0.0;
			    else if (pixel > maxval)
				scaled_pixel = MAXINT_F;
			    else
				scaled_pixel =(pixel - minval) *
				    intensity_scale;
			  }
			  else {	/* negative slope case */
			    if (pixel < maxval)
				scaled_pixel = MAXINT_F;
			    else if (pixel > minval)
				scaled_pixel = 0.0;
			    else
				scaled_pixel =(pixel - minval) *
				    intensity_scale;
			  }
			  byte = ((int)(scaled_pixel + 0.5)) >> 1;
			}
			if (colorMapping[byte] == 0xffffffff)
			{
				xcolor.red = byte << 8;
				xcolor.green = xcolor.red;
				xcolor.blue = xcolor.red;
				xcolor.flags = DoRed|DoGreen|DoBlue;
				XAllocColor(d,cmap,&xcolor);
				colorMapping[byte] = xcolor.pixel;
			}
			for (ipixrow=0;ipixrow<vpixsize;ipixrow++) {
				for (ipixcol=0;ipixcol<hpixsize;ipixcol++) {
					XPutPixel(ximage,xim_x_coord+ipixcol,xim_y_coord+ipixrow,colorMapping[byte]);
				}
			}
		}
	}
	Incr_Curcol += numcols;
	if (Incr_Curcol > XSize)
	{
		Incr_Currow += numrows;
		Incr_Curcol = 0;
	}
	if (Incr_refresh)
	{
		XPutImage(Incr_display,Incr_drawable,Incr_gc,ximage,
			  Incr_PutImParams.isrcx,Incr_PutImParams.isrcy,
			  Incr_PutImParams.idestx,Incr_PutImParams.idesty,
			  Incr_PutImParams.imw,Incr_PutImParams.imh);
	}
	return (Incr_Currow >= YSize);
}

Boolean Plane::Incr_ditherBitImageCallback()
{
	int xim_y_coord, xim_x_coord, bitoff, bitbyte;
	register int maskrow, maskcol, saved_maskcol, saved_maskrow;
	register int ipixrow, ipixcol;
	int crow,ccol,numrows,stoprow,numcols,stopcol;
	int byte;
	double intensity_scale,pixel,scaled_pixel;
	register short int *maskrow_ptr;
	int hpixsize,vpixsize;
	unsigned char *row_base;
	int last_maskcol;

	hpixsize = Incr_hpixsize;
	vpixsize = Incr_vpixsize;
	intensity_scale = Incr_intensity_scale;
	saved_maskrow = maskrow = Incr_maskrow;
	numrows = XIMThresh / vpixsize;
	numcols = XIMThresh / hpixsize;
	stoprow = Incr_Currow + numrows;
	stopcol = Incr_Curcol + numcols;
	if (stoprow > YSize) stoprow = YSize;
	if (stopcol > XSize) stopcol = XSize;

	for (crow = Incr_Currow; crow < stoprow; crow++)
	{
		row_base = (crow * XSize * pixel_size) + Data1D;
		xim_y_coord = crow * vpixsize;
		saved_maskrow = (maskrow % 16);
		
		for (ccol = Incr_Curcol; ccol < stopcol; ccol++)
		{
			maskcol = Incr_maskcol;
			xim_x_coord = ccol * hpixsize;
			switch (plane_type)
			{
				case LLVS_PLF_BYTE:
					pixel = (double) *(row_base + ccol);
					break;
				case LLVS_PLF_INT:
					pixel = (double) *(((long int *)row_base) + ccol);
					break;
				case LLVS_PLF_FLOAT:
					pixel = (double) *(((float *)row_base) + ccol);
					break;
			}
			/* scale pixel.  out of range values get black or white.
			   otherwise linear mapping */
			if (intensity_scale > 0.0) {
			    /* positive slope case */
			    if (pixel < minval)
				scaled_pixel = 0.0;
			    else if (pixel > maxval)
				scaled_pixel = MAXINT_F;
			    else
				scaled_pixel =(pixel - minval) *
				    intensity_scale;
			}
			else {	/* negative slope case */
			    if (pixel < maxval)
				scaled_pixel = MAXINT_F;
			    else if (pixel > minval)
				scaled_pixel = 0.0;
			    else
				scaled_pixel =(pixel - minval) *
				    intensity_scale;
			}
			byte = (int)(scaled_pixel + 0.5);
			saved_maskcol = (maskcol % 16);
			for (ipixrow=0;ipixrow<vpixsize;ipixrow++) {
				bitbyte = 1 << (bitoff & 0x07);
				if (maskrow > 15) maskrow = 0;
				maskrow_ptr = (short int *)(&matrix[maskrow][0]);
				maskcol = saved_maskcol;
				for (ipixcol=0;ipixcol<hpixsize;ipixcol++) {
					if (maskcol > 15) maskcol = 0;
					if (maskrow_ptr[maskcol] < byte) {
						XPutPixel(ximage,xim_x_coord+ipixcol,xim_y_coord+ipixrow,1);
					}/* else	XPutPixel(ximage,xim_x_coord+ipixcol,xim_y_coord+ipixrow,White);*/
					bitbyte = bitbyte << 1;
					if (bitbyte > 0x0080) {
						bitbyte = 1;
					}
					maskcol++;
				}
				maskrow++;
			}
		}
		last_maskcol = maskcol;
	}
	Incr_maskrow = maskrow;
	Incr_maskcol = last_maskcol;
	Incr_Curcol += numcols;
	if (Incr_Curcol > XSize)
	{
		Incr_Currow += numrows;
		Incr_Curcol = 0;
		Incr_maskcol = 0;
	}
	if (Incr_refresh)
	{
		XPutImage(Incr_display,Incr_drawable,Incr_gc,ximage,
			  Incr_PutImParams.isrcx,Incr_PutImParams.isrcy,
			  Incr_PutImParams.idestx,Incr_PutImParams.idesty,
			  Incr_PutImParams.imw,Incr_PutImParams.imh);
	}
	return (Incr_Currow >= YSize);
}


void Plane::SetColorMap(unsigned char rmap[],unsigned char gmap[],
		   unsigned char bmap[],int nc,Display *display,Colormap cmap)
{
	XColor xcolor;

	for (int byte=0;byte < nc;byte++)
	{
		xcolor.red = rmap[byte] << 8;
		xcolor.green = gmap[byte] << 8;
		xcolor.blue = bmap[byte] << 8;
		xcolor.flags = DoRed|DoGreen|DoBlue;
		XAllocColor(display,cmap,&xcolor);
		colorMapping[byte] = xcolor.pixel;
	}
	DontScale = True;
}



/**** helper code (Binary Plane I/O guts) *****/

/*
 * swap the bytes in an array of long words
 */

static void swap_longs(llvs_integer array[],int element_count)
{
    union {
	llvs_ubyte bytes[4];
	llvs_integer longword;
	} a,b;
    register int i;

    for (i = 0; i < element_count; i++) {
	a.longword = array[i];
	b.bytes[0] = a.bytes[3];
	b.bytes[1] = a.bytes[2];
	b.bytes[2] = a.bytes[1];
	b.bytes[3] = a.bytes[0];
	array[i] = b.longword;
	}
    }

/*
 * swap the bytes in an array of short words
 */

static void swap_words(llvs_half_integer array[],int element_count)
{
    union {
	llvs_ubyte bytes[2];
	llvs_half_integer word;
	} a,b;
    register int i;

    for (i = 0; i < element_count; i++) {
	a.word = array[i];
	b.bytes[0] = a.bytes[1];
	b.bytes[1] = a.bytes[0];
	array[i] = b.word;
	}
    }

/*
 * convert an array of floating point numbers from one format to
 * another format
 */

static void vax_to_ieee(unsigned char *cp,unsigned long int *ip);
static void ieee_to_vax(unsigned long int* ip,unsigned char* cp);

static void cvt_floats(llvs_integer array[],int element_count,
		       int current_format,int desired_format)
{
    static llvs_integer old,neww;
    register int i;

    for (i = 0; i < element_count; i++) {
	old = array[i];
	neww = 0;
/*
	if (LLVS_NATIVE_BYTE_SEX == LLVS_LOW_BYTE_FIRST && 
	     current_format == LLVS_IEEE_SINGLE_FLOAT) swap_longs(&old,1);
*/
	switch (current_format) {
	    case LLVS_DEC_SINGLE_FLOAT:
		switch (desired_format) {
		    case LLVS_DEC_SINGLE_FLOAT: neww = old; break;
		    case LLVS_IEEE_SINGLE_FLOAT:
			vax_to_ieee((unsigned char*)&old,
				    (unsigned long int*)&neww);
			break;
		    }
		break;
	    case LLVS_IEEE_SINGLE_FLOAT:
		switch (desired_format) {
		    case LLVS_DEC_SINGLE_FLOAT:
			ieee_to_vax((unsigned long int*)&old,
				    (unsigned char*) &neww);
			break;
		    case LLVS_IEEE_SINGLE_FLOAT: neww = old; break;
		    }
		break;
	    }
	if (LLVS_NATIVE_BYTE_SEX == LLVS_LOW_BYTE_FIRST &&
	    desired_format == LLVS_IEEE_SINGLE_FLOAT)
		swap_longs((long int *)&neww,1);
	array[i] = neww;
	}
    }

/*
 * VAX (also PDP-11) floating point to/from IEEE Single precision floating
 * point conversion functions.  Hairy magic here.  Only used if a floating
 * point LLVS plane actually writting on the VAX is read by a non-VAX (or
 * PDP-11).  Or if this code gets compiled on a VAX and reads a non-VAX FP
 * LLVS plane. I got this code from Jim Burrill.  I don't know where he
 * got it from.  The original was in VAX FORTRAN.
 */

/*****************************************************************
 fjm -- 11/6/87 -- Function IEEE_VAX_S(Value)

 This function takes an IEEE format MC68881 Single precision floating
 point number and returns the VAX F_floating form of the number if it
 is in range.

 Note: The smallest vax floating point number is 4 times smaller than
       the smallest ieee floating point number in normallized form.
       The largest vax floating point number is .5 times as large as
       the largest ieee floating point number.

       Actual ranges:

               Vax                     IEEE
               ---                     ----

       max     1.7014119x10**38        3.4028243x10**38
 
       min     0.29387365x10**-38      1.175495x10**-38 (normalized)
                                       1.4013x10**-45   (de-normalized)

 Let E = the exponent part of the real (in signed magnitude form), and
 let M = the mantisa, and S = the sign.

 The vax represents floating point numbers as:

       S bit -- bit 15
       E     -- bits 7-14, 8 bit, excess 128 exponent.
       M     -- bits 6(MSB)-0 and bits 31-16(LSB) unsigned fraction.
       Numbers are stored in a normalized form with a hidden 1 bit
       representing a 0.1 is immediately to the left of the MSB of M. Note
       that the radix point of the fraction is to the left of the hidden
       bit in this form.

       IF E .ge. 1 then
               value = (-1)**S x 2**(E-128) x 0.1M
       IF S .eq. 0 .and. E .eq. 0 then
               value = 0 (unsigned)
       IF S .eq. 1 .and. E .eq. 0 then
               illegal value, signalled on attempted operation.

 The IEEE representation is as follows:

       S bit -- bit 31
       E     -- bits 23-30, 8 bit, excess 127 binary exponent.
       M     -- bits 22(MSB) - 0(LSB) unsigned fraction.
       Numbers are stored in either normalized or de-normalized form. In
       normalized form (the usual representation) a hidden 1 bit representing
       a 1.0 is immediatly to the left of the MSB of M. In denormalized form,
       the hidden bit is 0. See below. Note that in IEEE form, the radix point
       of the fraction is to the right of the hidden bit.

       IF E .ge. 1 .and. E .le. 254 then
               actual value = (-1)**S x 2**(E-127) x 1.M
       IF E .eq. 0 .and. M .eq. 0 then
               value = signed zero (S bit is preserved)
       IF E .eq. 255 .and. M .ne. 0 then
               not a number (NAN). M can have any bit pattern, and the
               result of any operation on a NAN with the MSB of F set
               on an IEEE implementation results in a signal. The
               interpretation on any value associated with a NAN is left
               to the application.
       IF E .eq. 255 .and. M .eq. 0 then
               value = Plus or minus infinity, depending on S.
       IF E .eq. 0 .and. M .ne. 0 then
               denormalized number, used to represent numbers less 
               than can be represented in normalized form. The hidden bit
               is assumed to a zero.
               value = (-1)**S x 2**(E-126) x 0.M

 Note that the vax and ieee forms differ by a bias of 1 in their stored
 exponent form, and that they also differ implicitly by a power of 2 due to
 the difference in normalization scalling (ieee = 1.M, vax = 0.1M).
 Thus the conversion for the exponent is Evax = Eieee + 2. Further, the vax
 can represent smaller numbers in normallized form than can the 68881, so
 this routine converts the de-normalized numbers that are representable
 on the vax. Thus, all ieee normalized numbers with exponents less than 254
 are convertable to vax representations. Further, Some of the de-normalized
 numbers are convertable. Loss of range (overflow) occurs on the high end.
 The actual conversion outline follows:

 Normallized numbers:

       numbers in the magnitude of:
               1.175495 x 10**-38 to 1.7014119 x 10**38
       are converted exactly, with no loss of precision.

       numbers with a magnitude greater than 1.7014119 x 10**38
       are converted to a vax reserved operand with sign of 1 and
       exponent and mantissa of zero. This is the equivalent of
       the result of a floating overflow on a 780 class machine.

 De-normalized numbers:

       numbers in the magnitude of:
               0.29387365 x 10**-38 to 1.175494 x 10**-38
       are converted. Note that while the conversion is exact,
       the vax has 1 to 2 bits more precision in this range than is
       represenatble in the ieee form. The extra 1 or 2 bits are assigned
       zeros in the conversion.

       numbers less than 0.29387365 x10**-38 are mapped on to vax zero
       with S=0, E=0, and M=0.

 Zeros:

       both positive and minus zero in the ieee implementation are
       mapped on to vax zero with S=0, M=0, and E=0.

 Infinities:

       Infinities have no dual in the vax architecture, as such, the
       value mapped is a vax reserved operand, with S=1, E=0, and M=0.
       This is the equivalent of a floating overflow result on the
       780 class machines.

 Not Numbers (NAN's):

       These are mapped onto vax operands by setting the sign bit,
       clearing the exponent, and preserving the mantissa. Note that
       this conversion losses the original sign bit. The ieee representation
       has twice as many NAN's as the vax, so half are lost.

 
        integer*4 function IEEE_VAX_S(value) ! really of type bit vector.

        implicit        none

        integer*4       value

        integer*4       S
        integer*4       E
        integer*4       M
 Fetch the fields from IEEE format:

       31 30          23 22                           0
      +--+--------------+------------------------------+
      | S| Exponent     |MSB   Mantissa             LSB|
      +--+--------------+------------------------------+

        S = lib$extzv(31,1,value)
        E = lib$extzv(23,8,value)
        M = lib$extzv(0,23,value)

 Note: The following code is optimized to deal with the two most likely
       casses first: namely, normal values and zeros.


 Zeros and de-normalized numbers:
 
        if ( E .eq. 0 ) then

 Zeros:
          if( M .eq. 0 ) then
            S = 0

 de-normalized number:
 
          else ! M .ne. 0

          number is representable if  bit 21 or 22 is set.
          shift the mantisa the right number of bits and strip the
          high order set bit which becomes the hidden bit on the vax.
          Numbers outside the representable range are set to zero.

            if ( ( M .and. '400000'x) .ne. 0 ) then
              M = (M .or. '3FFFFF'x) * 2
              E = 2

            else if ( ( M .and. '200000'x) .ne. 0 ) then
              M = (M .or. '1FFFFF'x) * 4
              E = 1
            else
              S = 0
              E = 0
              M = 0
              end if
            end if

 Normal value:
 
        else if ( E .lt. 254 ) then
          E = E + 2


 un-representable (overflow)

        else if ( E .eq. 254 ) then
 

          S = 1
          E = 0
          M = 0

 Infinities and NAN's

        else ! E .eq. 255
          S = 1
          E = 0

          end if

 Store the numbers in vax format.

       31                   16 15 14       7 6           0
      +-----------------------+--+----------+-------------+
      |    Mantissa        LSB| S| Exponent |MSB Mantissa |
      +-----------------------+--+----------+-------------+


 
        call lib$insv(S,15,1,ieee_vax_s)
        call lib$insv(E, 7,8,ieee_vax_s)
        call lib$insv(lib$extzv(16,7,M),0,7,ieee_vax_s)
        call lib$insv(lib$extzv(0,16,M),16,16,ieee_vax_s)

        return
        end

    Frank J. Manion
                                        The Fox Chase Cancer Center
  Phone: (215) 728-3660                 7701 Burholme Avenue
                                        Philadelphia, PA  19111
                                        USA
*******************************************************************/

static void ieee_to_vax(unsigned long int* ip,unsigned char* cp)
{
int s, e, m;
unsigned int i;

/* Fetch the fields from IEEE format:

      31  30          23 22                           0
      +--+--------------+------------------------------+
      | S| Exponent     |MSB   Mantissa             LSB|
      +--+--------------+------------------------------+
*/

  i = *ip;
#if LLVS_NATIVE_BYTE_SEX == LLVS_LOW_BYTE_FIRST
/*  swap_longs(cp,1); */
  swap_longs((llvs_integer*)cp,1);
#endif
  s = 1 & (i >> 31);
  e = 0x0ff & (i >> 23);
  m = 0x7fffff & i;

/*
 Note: The following code is optimized to deal with the two most likely
       casses first: namely, normal values and zeros.
*/

 
  if (e == 0) { /* Zeros and de-normalized numbers: */
    if (m == 0) s = 0;
    else {
      /*
          number is representable if  bit 21 or 22 is set.
          shift the mantisa the right number of bits and strip the
          high order set bit which becomes the hidden bit on the vax.
          Numbers outside the representable range are set to zero.
      */
      if (m & 0x400000) {
        m = (m | 0x3FFFFF) << 1;
        e = 2;
        }
      else if (m & 0x200000) {
        m = (m | 0x1FFFFF) << 2;
        e = 1;
        }
      else {
        s = 0;
        e = 0;
        m = 0;
        }
      }
    }
  else if (e < 254) e += 2; /*  Normal value: */
  else if (e == 254) { /* un-representable (overflow) */
    s = 1;
    e = 0;
    m = 0;
    }
  else { /* E .eq. 255 - Infinities and NAN's */
    s = 1;
    e = 0;
    }

/*
  Store the numbers in vax format.

       31                   16 15 14       7 6           0
      +-----------------------+--+----------+-------------+
      |    Mantissa        LSB| S| Exponent |MSB Mantissa |
      +-----------------------+--+----------+-------------+
*/

  *cp++ = (m >> 16) | ((1 & e) << 7);
  *cp++ = (s << 7) | (e >> 1);
  *cp++ = 0xff & m;
  *cp++ = 0xff & (m >> 8);
}

void vax_to_ieee(unsigned char* cp,unsigned long int* ip)
{
int s, e, m;

/* Fetch the fields from VAX format:

       31                   16 15 14       7 6           0
      +-----------------------+--+----------+-------------+
      |    Mantissa        LSB| S| Exponent |MSB Mantissa |
      +-----------------------+--+----------+-------------+
*/

#if LLVS_NATIVE_BYTE_SEX == LLVS_HIGH_BYTE_FIRST
  swap_longs((llvs_integer*)cp,1);
#endif
  s = *(cp + 1) >> 7;
  e = ((0x7f & *(cp + 1)) <<  1) | (*(cp + 0) >> 7);
  m = ((0x7f & *(cp + 0)) << 16) | (*(cp + 3) << 8) | *(cp + 2);


/*
 Note: The following code is optimized to deal with the two most likely
       casses first: namely, normal values and zeros.
*/

 
  if (e > 1) e -= 2; /*  Normal value: */
  else if (e == 0) { /* Zeros and de-normalized numbers: */
    if (s == 0) m = 0;
    else { /* NAN & Infinity */
      m = 0;
      e = 255;
      }
    }
  else { /* un-representable (overflow) */
    s = 0;
    e = 0;
    m = 0;
    }
/*
  Store the numbers in IEEE format.

       31 30          23 22                           0
      +--+--------------+------------------------------+
      | S| Exponent     |MSB   Mantissa             LSB|
      +--+--------------+------------------------------+
*/

  *ip = (s << 31) | (e << 23) | m;

#if LLVS_NATIVE_BYTE_SEX == LLVS_LOW_BYTE_FIRST
  swap_longs((llvs_integer*)ip,1);
#endif
}


