/*
   ColorImage object can convert from Red, Green & Blue to Hue, Saturation 
   and Value (and back again). It is also able to convert from RGB to
   intensity (but obviously, not back again).
   Programmer: Kirill Yevseyev
   $Id: ColorImage.h,v 1.5 1994/11/14 17:40:05 heller Exp $
   $Log: ColorImage.h,v $
 * Revision 1.5  1994/11/14  17:40:05  heller
 * Fix writeGifFile's protection (does not belong under #ifdef TIFFLIB)
 *
 * Revision 1.4  1994/10/26  16:39:06  heller
 * Added ColorImage construction that takes just size info (creates zeroed
 * planes).
 *
 * Revision 1.3  1994/10/17  18:08:06  heller
 * Added support for writing viff 3-band color images.
 *
 * Revision 1.2  1994/10/14  19:45:11  heller
 * Added support for reading viff 3-band color images.
 *
*/

#ifndef COLOR_IMAGE_DEFINED
#define COLOR_IMAGE_DEFINED
#include <Token.h>
#include <Isrstream.h>
#include <NewPlane.h>


class ColorImage : public Token {
 public:
  /* Constructor */
  ColorImage ();
  ColorImage (int XSize, int YSize, int XOrigin, int YOrigin,
	      AxisDirection XDir = Up, AxisDirection YDir = Up,
	      Boolean RGB = True);
  ColorImage (FloatPlane *hue, FloatPlane *saturation, FloatPlane *value);
  ColorImage (BytePlane *red, BytePlane *green, BytePlane *blue);
  /* Destructor */
  ~ColorImage ();

  char *name() {return "ColorImage";}
  void input(Isristream& is); 
  void output(Isrostream& os);
  void trace();

  void BitDraw (Display* display, Drawable drawable, GC gc,
		Transform2D* trans, float xmin, float ymin,
		float xmax, float ymax, Visual *visual,
		unsigned int depth, Colormap cmap);

  BytePlane  *red () {if (red_plane == NULL) compute_red_green_blue ();
		      return red_plane;}
  BytePlane  *green () {if (green_plane == NULL) compute_red_green_blue ();
			return green_plane;}
  BytePlane  *blue () {if (blue_plane == NULL) compute_red_green_blue ();
		       return blue_plane;}
  FloatPlane *hue () {if (hue_plane == NULL) compute_hue_saturation_value ();
		      return hue_plane;}
  FloatPlane *saturation () {if (saturation_plane == NULL) 
			       compute_hue_saturation_value ();
			     return saturation_plane;}
  FloatPlane *value () {if (value_plane == NULL) 
			  compute_hue_saturation_value ();
			return value_plane;}
  BytePlane  *intensity () {if (intensity_plane == NULL)
			      compute_intensity ();
			    return intensity_plane;}

  void SetRedGreenBlue (BytePlane *r, BytePlane *g, BytePlane *b);
  void SetHueSaturationValue (FloatPlane *h, FloatPlane *s, FloatPlane *v); 

  CoordSys2D *coordinates() {return value ()->coordinates();}

  friend void copy_ColorImage_header(ColorImage* source,ColorImage* dest);
  friend void writeGifFile(char *filename,ColorImage *cimg);
#ifdef TIFFLIB
  friend void writeTiffFile(char *filename,ColorImage *cimg);
  friend Token* ReadTiffFile(char *filename);
#endif
#ifdef KHOROS
  friend ColorImage* viff2ColorImage(struct xvimage *image);
  xvimage *viff();
#endif
 private:
  BytePlane  *red_plane;
  BytePlane  *green_plane;
  BytePlane  *blue_plane;
  FloatPlane *hue_plane;
  FloatPlane *saturation_plane;
  FloatPlane *value_plane;
  BytePlane  *intensity_plane;
  
  int compute_red_green_blue (); /* if hue, saturation, and value are defined,
				    compute and store red, green, and blue, 
				    and return 1. Otherwise, return -1. */
  int compute_hue_saturation_value ();  /* if red, green, and blue are defined,
					   compute and store hue, saturation,
					   and value, and return 1. Otherwise, 
					   return -1. */
  void compute_intensity (); /* computes and store intensity. */

  BytePlane  *map_plane;	/* 8-bit color mapped image */
  unsigned char rmap[256],gmap[256],bmap[256]; /* color maps */
  void MakeMapPlane(int ncolors=256,int bits=8,double fraction=1.0);

  struct pixel
  {
/* The following line replaces the next one */
    unsigned int offset;
/*    unsigned short x,y;   */                /* Location of pixel */
    pixel *next;                   /* Pointer to next one */
  };
  struct color
  {
    unsigned char r;                      /* Color values */
    unsigned char g;
    unsigned char b;
    int clr;                              /* Packed color */
    int count;                            /* Pixel count */
    color *prev;                   /* Ptrs to previous & next color */
    color *next;
    pixel *pixlist;                /* List of pixels of this color */
    pixel *pixtail;                /* End of pixel list */
  };
  struct box
  {
    unsigned char rmin,rmax;              /* Bounds on box */
    unsigned char gmin,gmax;
    unsigned char bmin,bmax;
    color *colors;                 /* List of colors in box */
    box *next;                     /* Next box */
    int count;                            /* Number of colors in box */
  };
  void load_color_list();
  box *find_biggest_box2();
  box *find_biggest_box1();
  void box_move_color(box *b1,box *b2,color **p1,color **p2);
  void compress_colors(int *ncolors,float fraction);
  void gamut_head_swap(color *p);
  color *cmalloc();
  void gamut_adjust_image(int ncolors);
  void gamut_set_box_bounds(box *b);
  void free_colors();
  void sort_cmaps(int nc);
};
  
#endif






