/**************************************************************************
 * Isrostream & Isristream classes                                        *
 *    Isrostream and Isristream are output and input streams for ISR3.1   *
 * tokens. Every write to an Isrostream writes a token, plus all the      *
 * tokens it points to (recursively). Every read from an Isristream reads *
 * one token and any tokens it points to (recursively). Standard use (at  *
 * least for files) is to open a stream, write/read one token that points *
 * to others (such as a TokenSet), and then close the stream. Circular    *
 * token pointers are not a problem. The Isrostream and Isristream classes*
 * can only be used to write ISR3.1 tokens for which the input(), output()*
 * trace() and name() virtual methods are defined. See Token.h for a      *
 * description of these functions.                                        *
 *                                                                        *
 *    Right now, Isrostream writes to a file and Isristream reads from a  *
 * file, but in the future direct IPC will be an option.                  *
 *                                                                        *
 *   There are two ways to read a token. You can create an instance of the*
 * appropriate class, and fill it using the C++ ">>" operator. Or, if the *
 * class type is unknown, the function read_token_ptr(stream) will read   *
 * any type of ISR3.1 token and return a pointer to it.                   *
 *                                                                        *
 *   To write a token, use the C++ "<<" operator.                         * 
 *                                                                        *
 *   Streams have two user-settable properties: format and type. The      *
 * format of a stream is ASCII or BINARY. It determines whether ints,     *
 * floats, doubles, shorts and longs are written in their binary          *
 * representation or in ascii. Token pointers are also more compact in    *
 * binary than in ASCII. Byte-order (Big vs Little Endian) is taken care  *
 * of for binary streams. Binary is the default, since it is more compact,*
 * and binary files can easily be translated to ASCII. ASCII files can be *
 * useful for debugging, however. WARNING: In the BINARY format,          *
 * Isrostreams do not print character strings. This allows output         *
 * functions like:                                                        *
 *    ostream << "field name " << field;                                  *
 * that print labels when called on an ASCII stream but are compact for   *
 * BINARY streams. If a character string is data and you need it stored   *
 * in binary files, use the print_string function. NOTE: in ASCII mode,   *
 * reading in a constant string such as:                                  *
 *   istream >> "field name " >> field;                                   *
 * double checks that the string is in the file, and then skips over it.  *
 * The two lines of code above therefore complement each other. In        *
 * BINARY mode, reading into a constant string is a no-op.                *
 *                                                                        *
 *   The type of a stream is ISR or FORIEGN. ISR3.1 streams have headers  *
 * that set up data type tables and such. All pointers will be correctly  *
 * de-referenced in ISR3.1 files. FORIEGN files have unknown headers and  *
 * formats; the presumption is that the input() and output() functions    *
 * know how to read/write the foriegn format, and that there are no       *
 * pointers to other tokens (or if there are, the input() and output      *
 * functions deal with it). read_token_ptr() is not valid for foriegn     *
 * files; the user must allocate a token and use ">>" to read.            *
 *                                                                        *
 *   A handy function match_stream_to_filename is provided to match the   *
 * format and type of a stream to common filename conventions.            *
 *************************************************************************/


#ifndef ISRSTREAM_DEFINED
#define ISRSTREAM_DEFINED

#include<iostream.h>
#include<fstream.h>
#include<iomanip.h>
#include<stdlib.h> 
#include<assert.h>
#include<ctype.h>
#include<string.h>

/* stream types */
#define UNTYPED -1
#define BINARY 0
#define ASCII  1
#define ISR 0
#define FORIEGN 1
#define VIFF 2

#define MAX_TOKENTYPE_NAME_LENGTH 50

#ifndef ISR_TRUE
#define ISR_TRUE 1
#endif
#ifndef ISR_FALSE
#define ISR_FALSE 0
#endif

class TypeDirectory;
enum StreamStatus {Open, Closed};

class Isrstream {
 public:
  int format;         /* BINARY or ASCII */
  char *name;         /* filename associated with stream */
 protected:
  int type;           /* ISR  */
  StreamStatus state; /* Open or Closed */
};

class Isrostream : public Isrstream {
  friend class ISROStream;
  friend Isrostream& operator << (Isrostream&, const short);
  friend Isrostream& operator << (Isrostream&, const int);
  friend Isrostream& operator << (Isrostream&, const long);
  friend Isrostream& operator << (Isrostream&, const unsigned short);
  friend Isrostream& operator << (Isrostream&, const unsigned int);
  friend Isrostream& operator << (Isrostream&, const unsigned long);
  friend Isrostream& operator << (Isrostream&, const float);
  friend Isrostream& operator << (Isrostream&, const double);
  friend Isrostream& operator << (Isrostream&, const char*);
  friend Isrostream& operator << (Isrostream&, const char);
  friend int operator! (Isrostream&);

 public:
  Isrostream(char *filename, int stream_format = BINARY, int stream_type = ISR);
  Isrostream(int stream_format = BINARY, int stream_type = ISR);;
  ~Isrostream();

  void print_string(char *);
  Isrostream& write(const char *s, int n);

  void write_magic_header();

 private:
  void open(char *, int);
  void close();

  ofstream *str;
};

/* These should all be done with a template
   function, but I have had som many troubles
   with internal compiler errors and getting
   the linker to handle seperate compilation
   with template functions that I have decided
   to use brute force instead.                 */

Isrostream& operator << (Isrostream&, const int);
Isrostream& operator << (Isrostream&, const long);
Isrostream& operator << (Isrostream&, const short);
Isrostream& operator << (Isrostream&, const unsigned int);
Isrostream& operator << (Isrostream&, const unsigned long);
Isrostream& operator << (Isrostream&, const unsigned short);
Isrostream& operator << (Isrostream&, const float);
Isrostream& operator << (Isrostream&, const double);
Isrostream& operator << (Isrostream&, const char*);
Isrostream& operator << (Isrostream&, const char);
int operator! (Isrostream&);

#define FileVersion 1
#define MagicASCII "ISR3.1XY"
#define MagicB1    "ISR3.1"
#define MagicB2    0x5859

class Isristream : public Isrstream {
  friend class ISRIStream;
  friend Isristream& operator >> (Isristream&,  short&);
  friend Isristream& operator >> (Isristream&,  unsigned short&);
  friend Isristream& operator >> (Isristream&,  int&);
  friend Isristream& operator >> (Isristream&,  unsigned int&);
  friend Isristream& operator >> (Isristream&,  long&);
  friend Isristream& operator >> (Isristream&,  unsigned long&);
  friend Isristream& operator >> (Isristream&,  float&);
  friend Isristream& operator >> (Isristream&,  double&);
  friend Isristream& operator >> (Isristream&,  const char*);
  friend Isristream& operator >> (Isristream&,  const char);
  friend int operator! (Isristream&);

 public:
  TypeDirectory *tokentype_table;

  char *read_string(int);
  Isristream& read(char *ptr, int n);
  void putback(char);
  char get();

  Isristream(char *filename, int stream_format = BINARY, int stream_type = ISR);
  Isristream(int stream_format = BINARY, int stream_type = ISR);
  ~Isristream();

  void read_magic_header();

 private:
  void open(char *, int mode);
  void close();

  void read_tokentype_table(int, int);
  int read_cstring(const char *cstring);

  ifstream *str;
  unsigned short int fileVersion, byteSwapNeeded;
};

Isristream& operator >> (Isristream&, short&);
Isristream& operator >> (Isristream&, unsigned short&);
Isristream& operator >> (Isristream&, int&);
Isristream& operator >> (Isristream&, unsigned int&);
Isristream& operator >> (Isristream&, long&);
Isristream& operator >> (Isristream&, unsigned long&);
Isristream& operator >> (Isristream&, float&);
Isristream& operator >> (Isristream&, double&);
Isristream& operator >> (Isristream&, const char*);
Isristream& operator >> (Isristream&, const char);
int operator! (Isristream&);

int match_stream_to_filename(Isrstream&, char*);
#endif /* ISRSTREAM_DEFINED */




