/*
 * ------------------------------------------------------------------
 * TiffIO.cc - Tiff I/O
 * Created by Robert Heller on Wed Aug 24 14:16:21 1994
 * ------------------------------------------------------------------
 * Modification History:
 * ------------------------------------------------------------------
 * Contents:
 * ------------------------------------------------------------------
 *  
 * 
 *            Copyright 1994 University of Massachusetts.
 *                          All rights reserved.
 * 
 *      Permission to copy and modify this software and its documen-
 *      tation only for internal use in your organization is hereby
 *      granted, provided that this notice is retained thereon and
 *      on all copies.  UMASS makes no representations as to the sui-
 *      tability and operability of this software for any purpose.
 *      It is provided "as is" without express or implied warranty.
 * 
 *      UMASS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 *      INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
 *      NESS.  IN NO EVENT SHALL UMASS BE LIABLE FOR ANY SPECIAL,
 *      INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY OTHER DAMAGES WHAT-
 *      SOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 *      IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 *      ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PER-
 *      FORMANCE OF THIS SOFTWARE.
 * 
 *      No other rights, including, for example, the right to redis-
 *      tribute this software and its documentation or the right to
 *      prepare derivative works, are granted unless specifically
 *      provided in a separate license agreement.
 * 
 *      Copyright 1994, University of Massachusetts. All rights
 *      reserved.
 * 
 * 
 */

#include <Token.h>
#include <NewPlane.h>
#include <ColorImage.h>
#include <tiffio.h>

#define MAXCOLORS 2028
#define COLIM_MAXVAL 255

void writeTiffFile(char *filename,BytePlane *bpl)
{
    unsigned char *rowptr;
    int cols, rows, row;
    TIFF* tif;
    unsigned short compression;
    unsigned short fillorder;
    short photometric;
    short samplesperpixel;
    short bitspersample;
    int bytesperrow;
    unsigned char* buf;

    bpl->GetSize(cols,rows);
    samplesperpixel = 1;
    bitspersample = 8;
    photometric = PHOTOMETRIC_MINISBLACK;
    bytesperrow = cols;
    compression = COMPRESSION_LZW;
    fillorder = FILLORDER_MSB2LSB;
    buf = new unsigned char[bytesperrow];
    if ( buf == NULL )
    {
    	cerr << "can't allocate memory for row buffer\n";
    	return;
    }    
    tif = TIFFOpen(filename,"w");
    if ( tif == NULL )
    {
    	cerr << "error opening " << filename << " as TIFF file\n";
	delete buf;
    	return;
    }

    /* Set TIFF parameters. */
    TIFFSetField( tif, TIFFTAG_IMAGEWIDTH, cols );
    TIFFSetField( tif, TIFFTAG_IMAGELENGTH, rows );
    TIFFSetField( tif, TIFFTAG_BITSPERSAMPLE, bitspersample );
    TIFFSetField( tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT );
    TIFFSetField( tif, TIFFTAG_COMPRESSION, compression );
    TIFFSetField( tif, TIFFTAG_PHOTOMETRIC, photometric );
    TIFFSetField( tif, TIFFTAG_FILLORDER, fillorder );
    TIFFSetField( tif, TIFFTAG_DOCUMENTNAME, bpl->Name);
    TIFFSetField( tif, TIFFTAG_IMAGEDESCRIPTION, "converted ISR3.1 BytePlane");
    TIFFSetField( tif, TIFFTAG_SAMPLESPERPIXEL, samplesperpixel );
    TIFFSetField( tif, TIFFTAG_ROWSPERSTRIP, ( 8 * 1024 ) / bytesperrow);
    TIFFSetField( tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG );
    /* Now write the TIFF data. */
    for ( row = 0,rowptr = bpl->GetDataPtr(); row < rows; ++row,rowptr += cols )
    {
	memcpy(buf,rowptr,cols);
	if ( TIFFWriteScanline( tif, buf, row, 0 ) < 0 )
	{
		cerr << "failed a scanline write on row " << row << "\n";
		delete buf;
		TIFFClose(tif);
		return;
	}
    }
    TIFFFlushData( tif );
    TIFFClose( tif );
}

void writeTiffFile(char *filename,ColorImage *cimg)
{
    unsigned char *redptr,*greenptr,*blueptr,*rpix,*gpix,*bpix,*tpix;
    int cols, rows, row, col;
    TIFF* tif;
    unsigned short compression;
    unsigned short fillorder;
    short photometric;
    short samplesperpixel;
    short bitspersample;
    int bytesperrow;
    unsigned char* buf;

    cimg->red()->GetSize(cols,rows);
    samplesperpixel = 3;
    bitspersample = 8;
    photometric = PHOTOMETRIC_RGB;
    bytesperrow = cols * 3;
    compression = COMPRESSION_LZW;
    fillorder = FILLORDER_MSB2LSB;
    buf = new unsigned char[bytesperrow];
    if ( buf == NULL )
    {
    	cerr << "can't allocate memory for row buffer\n";
    	return;
    }    
    tif = TIFFOpen(filename,"w");
    if ( tif == NULL )
    {
    	cerr << "error opening " << filename << " as TIFF file\n";
	delete buf;
    	return;
    }

    /* Set TIFF parameters. */
    TIFFSetField( tif, TIFFTAG_IMAGEWIDTH, cols );
    TIFFSetField( tif, TIFFTAG_IMAGELENGTH, rows );
    TIFFSetField( tif, TIFFTAG_BITSPERSAMPLE, bitspersample );
    TIFFSetField( tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT );
    TIFFSetField( tif, TIFFTAG_COMPRESSION, compression );
    TIFFSetField( tif, TIFFTAG_PHOTOMETRIC, photometric );
    TIFFSetField( tif, TIFFTAG_FILLORDER, fillorder );
    TIFFSetField( tif, TIFFTAG_DOCUMENTNAME, "");
    TIFFSetField( tif, TIFFTAG_IMAGEDESCRIPTION, "converted ISR3.1 ColorImage");
    TIFFSetField( tif, TIFFTAG_SAMPLESPERPIXEL, samplesperpixel );
    TIFFSetField( tif, TIFFTAG_ROWSPERSTRIP, ( 8 * 1024 ) / bytesperrow);
    TIFFSetField( tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG );
    /* Now write the TIFF data. */
    redptr = cimg->red()->GetDataPtr();
    greenptr = cimg->green()->GetDataPtr();
    blueptr = cimg->blue()->GetDataPtr();
    for ( row = 0; row < rows; ++row,redptr += cols,greenptr += cols,blueptr += cols )
    {
	for (col = 0,tpix = buf,rpix=redptr,gpix=greenptr,bpix=blueptr; col < cols; col++,rpix++,gpix++,bpix++)
	{
		*tpix++ = *rpix;
		*tpix++ = *gpix;
		*tpix++ = *bpix;
	}
	if ( TIFFWriteScanline( tif, buf, row, 0 ) < 0 )
	{
		cerr << "failed a scanline write on row " << row << "\n";
		delete buf;
		TIFFClose(tif);
		return;
	}
    }
    TIFFFlushData( tif );
    TIFFClose( tif );
    delete buf;
}

struct ColorMaping {
	unsigned char rpix,gpix,bpix;
};

Token* ReadTiffFile(char *filename)
{
    int cols, rows, grayscale;
    int numcolors;
    register TIFF* tif;
    int row, i;
    register int col;
    u_char* buf;
    register u_char* inP;
    int maxval;
    static ColorMaping colormap[MAXCOLORS];
    register u_char sample;
    register int bitsleft;
    unsigned short bps, spp, photomet;
    unsigned short* redcolormap;
    unsigned short* greencolormap;
    unsigned short* bluecolormap;
    BytePlane *redplane,*greenplane,*blueplane;
    unsigned char *redrow,*redpix,*greenrow,*greenpix,*bluerow,*bluepix;
    Token* result;
    
    tif = TIFFOpen(filename,"r");
    if ( tif == NULL )
    {
    	cerr << "error opening TIFF file " << filename << "\n";
    	return(NULL);
    }
    if ( ! TIFFGetField( tif, TIFFTAG_BITSPERSAMPLE, &bps ) )
	bps = 1;
    if ( ! TIFFGetField( tif, TIFFTAG_SAMPLESPERPIXEL, &spp ) )
	spp = 1;
    if ( ! TIFFGetField( tif, TIFFTAG_PHOTOMETRIC, &photomet ) )
    {
    	cerr << "error getting photometric\n";
	TIFFClose(tif);
    	return(NULL);
    }

    switch ( spp )
	{
	case 1:
	case 3:
	case 4:
	break;

	default:
	cerr << "can only handle 1-channel gray scale or 1- or 3-channel color\n";
	return(NULL);
	}

    (void) TIFFGetField( tif, TIFFTAG_IMAGEWIDTH, &cols );
    (void) TIFFGetField( tif, TIFFTAG_IMAGELENGTH, &rows );

    maxval = ( 1 << bps ) - 1;
    if ( maxval == 1 && spp == 1 )
	{
	grayscale = 1;
	}
    else
	{
	switch ( photomet )
	    {
	    case PHOTOMETRIC_MINISBLACK:
	    grayscale = 1;
	    break;

	    case PHOTOMETRIC_MINISWHITE:
	    grayscale = 1;
	    break;

	    case PHOTOMETRIC_PALETTE:
	    if ( ! TIFFGetField( tif, TIFFTAG_COLORMAP, &redcolormap, &greencolormap, &bluecolormap ) )
	    {
		cerr << "error getting colormaps\n";
		TIFFClose(tif);
		return(NULL);
	    }
	    numcolors = maxval + 1;
	    if ( numcolors > MAXCOLORS )
	    {
	    	cerr << "too many colors\n";
	    	TIFFClose(tif);
	    	return(NULL);
	    }
	    maxval = COLIM_MAXVAL;
	    grayscale = 0;
	    for ( i = 0; i < numcolors; ++i )
		{
		register unsigned char r, g, b;
		r = (int) ((redcolormap[i]/65535.0) * COLIM_MAXVAL);
		g = (int) ((greencolormap[i]/65535.0) * COLIM_MAXVAL);
		b = (int) ((bluecolormap[i]/65535.0) * COLIM_MAXVAL);
		colormap[i].rpix = r;
		colormap[i].gpix = g;
		colormap[i].bpix = b;
		}
	    break;

	    case PHOTOMETRIC_RGB:
	    grayscale = 0;
	    break;

	    case PHOTOMETRIC_MASK:
	    cerr << "don't know how to handle PHOTOMETRIC_MASK\n";
	    TIFFClose(tif);
	    return(NULL);


	    case PHOTOMETRIC_SEPARATED:
	    cerr << "don't know how to handle PHOTOMETRIC_SEPARATED\n";
	    TIFFClose(tif);
	    return(NULL);


	    case PHOTOMETRIC_YCBCR:
	    cerr << "don't know how to handle PHOTOMETRIC_YCBCR\n";
	    TIFFClose(tif);
	    return(NULL);


	    case PHOTOMETRIC_CIELAB:
	    cerr << "don't know how to handle PHOTOMETRIC_CIELAB\n";
	    TIFFClose(tif);
	    return(NULL);


	    default:
	    cerr << "unknown photometric: " << photomet << "\n";
	    TIFFClose(tif);
	    return(NULL);

	    }
	}
    if ( maxval > COLIM_MAXVAL )
    {
	cerr << "bits/sample is too large\n";
	TIFFClose(tif);
	return(NULL);
    }
    buf = new u_char[TIFFScanlineSize(tif)];
    if ( buf == NULL )
    {
    	cerr << "can't allocate memory for scanline buffer\n";
    	TIFFClose(tif);
    	return(NULL);
    }
    if ( grayscale )
    {
    	redplane = new BytePlane(cols,rows,0,0,Up,Up,"TIFF Grey");
	redrow = redplane->GetDataPtr();
    	greenplane = blueplane = NULL;
    	result = (Token*) redplane;
    } else
    {
    	ColorImage *ci = new ColorImage(cols,rows,0,0,Up,Up);
    	result = (Token*) ci;
    	redplane = ci->red();
	redrow = redplane->GetDataPtr();
    	greenplane = ci->green();
	greenrow = greenplane->GetDataPtr();
    	blueplane = ci->blue();
	bluerow = blueplane->GetDataPtr();
    }

#define NEXTSAMPLE \
    { \
    if ( bitsleft == 0 ) \
	{ \
	++inP; \
	bitsleft = 8; \
	} \
    bitsleft -= bps; \
    sample = ( *inP >> bitsleft ) & maxval; \
    }

    for ( row = 0; row < rows; ++row )
    {
	if ( TIFFReadScanline( tif, buf, row, 0 ) < 0 )
	{
		cerr << "bad data read on line " << row << "\n";
		TIFFClose(tif);
		if (((BytePlane*)result) == redplane)
		{
		  delete (BytePlane*)result;
		} else
		{
		  delete (ColorImage*)result;
		}
		delete buf;
		return(NULL);
	}
	inP = buf;
	bitsleft = 8;
	redpix = redrow;
	greenpix = greenrow;
	bluepix = bluerow;
	switch ( photomet )
	    {
	    case PHOTOMETRIC_MINISBLACK:
	    for ( col = 0; col < cols; ++col )
		{
		NEXTSAMPLE
		*redpix++ = sample;
		}
	    break;

	    case PHOTOMETRIC_MINISWHITE:
	    for ( col = 0; col < cols; ++col )
		{
		NEXTSAMPLE
		sample = maxval - sample;
		*redpix++ = sample;
		}
	    break;

	    case PHOTOMETRIC_PALETTE:
	    for ( col = 0; col < cols; ++col )
		{
		NEXTSAMPLE
		*redpix++ = colormap[sample].rpix;
		*greenpix++ = colormap[sample].gpix;
		*bluepix++ = colormap[sample].bpix;
		}
	    break;

	    case PHOTOMETRIC_RGB:
	    for ( col = 0; col < cols; ++col )
		{
		register unsigned char r, g, b;

		NEXTSAMPLE
		r = sample;
		NEXTSAMPLE
		g = sample;
		NEXTSAMPLE
		b = sample;
		if ( spp == 4 )
		    NEXTSAMPLE		/* skip alpha channel */
		*redpix++ = r;
		*greenpix++ = g;
		*bluepix++ = b;
		}
	    break;

	    default: break;
	    
	    }
	redrow += cols;
	greenrow += cols;
	bluerow += cols;
    }
    delete buf;
    TIFFClose(tif);
    return result;
}


