/* =============================================================
   mdtrain: Gathers training data for building a 
            multivariate decision tree for classification of
	    a color image into target and non-target regions.

	    Usage is as follows:
	      mdtrain training_image(input) feature_file(output)

	      'training_image' is the image from which the training
	      data (pixels) is to be extracted.  There are two types of 
	      training data -- positive (class 1), representing targets,
	      and negative (class 0), representing non-target regions.

	      After extraction of sub-regions, a 'feature-file' is built
	      by the program, which stores the training data in a form
	      usable by the routine which builds the decision tree.

   Shashi Buluswar
   Computer Vision Laboratory
   Dept of Computer Science
   University of Massachusetts
   Amherst, MA  01003

   Copyright 1995, University of Massachusetts - Amherst

   =============================================================== */

#define MAIN
#include <mdtrain.h>


/*************************************************************
* strcpycat(s1, s2)                                          *
*    A minor helper function that returns a pointer to a     *
* string which is a copy of s1 concatenate s2. This function *
* is not destructive.                                        *
*************************************************************/

char *strcpycat(char *s1, char *s2)
{
  char *newstring;

  newstring = malloc(sizeof(char) * (strlen(s1) + strlen(s2) + 1));
  strcpy(newstring, s1);
  strcat(newstring, s2);
  return(newstring);
}

void Write2FeatFile(char *source, int pix_class, float pct, char *dest)
{

  FILE *red, *green, *blue;
  FILE *output;

  char *redfilename, *greenfilename, *bluefilename;
  char dummystring[100];

  int redval, greenval, blueval, temp;

  /* expand source file name into source_r, source_g, and source_b */
  redfilename = strcpycat(source, "R.asc");
  greenfilename = strcpycat(source, "G.asc");
  bluefilename = strcpycat(source, "B.asc");

  /* open red, green and blue files */
  if (NULL == (red = fopen(redfilename, "r"))) {
    fprintf(stderr, "\nunable to open %s\n", redfilename);
    exit(-1);
  }
  if (NULL == (green = fopen(greenfilename, "r"))) {
    fprintf(stderr, "\nunable to open %s\n", greenfilename);
    exit(-1);
  }
  if (NULL == (blue = fopen(bluefilename, "r"))) {
    fprintf(stderr, "\nunable to open %s\n", bluefilename);
    exit(-1);
  }

  /* Skip over header info put in ascii VIFF files */
  for (temp = 0; temp < 14; temp++) {
    fgets(dummystring,100,red);
    fgets(dummystring,100,green);
    fgets(dummystring,100,blue);
  }

  /* open output file for APPEND */
  if (NULL == (output = fopen(dest, "a"))) {
    fprintf(stderr, "\nunable to open %s for append\n", dest);
    exit(-1);
  }

  /* sample source pixels into output file */
  while ((temp = fscanf(red, "[%*d][%*d] = %d    ", &redval))
	 != EOF) {
    if (temp == 0) {
      fprintf(stderr, "\nRed plane is in an incorrect format\n");
      exit(-1);
    }

    temp = fscanf(green, "[%*d][%*d] = %d    ", &greenval);
    if (temp == EOF) {
      fprintf(stderr, "Green Plane is smaller then Red!\n");
      exit(-1);
    }
    else if (temp == 0) {
      fprintf(stderr, "\nGreen plane is in an incorrect format\n");
      exit(-1);
    }
  
    temp = fscanf(blue, "[%*d][%*d] = %d    ", &blueval);
    if (temp == EOF) {
      fprintf(stderr, "Blue Plane is smaller then Red!\n");
      exit(-1);
    }
    else if (temp == 0) {
      fprintf(stderr, "\nBlue plane is in an incorrect format\n");
      exit(-1);
    }


    if (pct >= frandom())
      fprintf(output, "%d %d %d %d\n", pix_class, redval, greenval, blueval);
  }
  fclose(output);
  fclose(red);
  fclose(green);
  fclose(blue);
}

/* Convert input image to viff, for use with EditImage, a Khoros routine */

void ConvertInImg2Viff()
{

  InStream.MatchFormatToFilename(IN_IMG);
  InStream.open(IN_IMG);
  if (!InStream) {
    cerr << "Unable to open file " <<  IN_IMG << endl;
    exit(-1);
  }

  InImgData = InStream.read_token_ptr();

  TmpStream.MatchFormatToFilename(TMP_IMG);
  TmpStream.open(TMP_IMG);
  if (!TmpStream) {
    cerr << "Unable to open file " <<  TMP_IMG << endl;
    exit(-1);
  }

  TmpStream << *InImgData; 
}


void CreateFeatFromImg()
{

  FILE *tmpfp;
  int i=0, tmpClass;
  strtype fViff, fViff0, fViff1, fViffR, fViffG, fViffB, cmd, fAsc, fAscR, fAscG, fAscB;

  /* === pick out all mdTMP#.viff RGB files and split into R,G,B === */
  /* === then convert individual R,G,B viff files into ascii     === */
  /* === thereafter, create feature file                         === */

  sprintf (fViff0, "mdTMP%d.viff0", i); 
  sprintf (fViff1, "mdTMP%d.viff1", i); 
  sprintf (fViffR, "mdTMP%dR.viff", i);   
  sprintf (fViffG, "mdTMP%dG.viff", i); 
  sprintf (fViffB, "mdTMP%dB.viff", i); 
  
  while (((tmpfp = fopen (fViff0, "r")) != NULL) || 
	 ((tmpfp = fopen (fViff1, "r")) != NULL)) {

    /* === check subimg pixel class === */

    if ((tmpfp = fopen (fViff0, "r")) != NULL) {
      fViff = fViff0;
      tmpClass = 0;
    }
    else if ((tmpfp = fopen (fViff1, "r")) != NULL) {
      fViff = fViff1;
      tmpClass = 1;
    }
    
    /* === split subimg into R,G,B bands using khoros vbandspt3 === */
    /* === and get rid of tmp files === */

    printf ("reading sub-image: %d...\n", i);
    sprintf (cmd, "vbandspt3 -i %s -o1 %s -o2 %s -o3 %s", fViff, fViffR, fViffG, fViffB);
     system (cmd);
    
    sprintf (cmd, "rm %s", fViff);
    system (cmd);
    
    /* === convert viff files to ascii using khoros vprdata === */
    /* === and get rid of tmp files === */

    printf ("creating ascii...\n");
    sprintf (fAscR, "mdTMP%dR.asc", i);   
    sprintf (cmd, "vprdata -i %s -f %s", fViffR, fAscR);
    system (cmd);
    
    sprintf (cmd, "rm %s", fViffR);
    system (cmd);
    
    sprintf (fAscG, "mdTMP%dG.asc", i); 
    sprintf (cmd, "vprdata -i %s -f %s", fViffG, fAscG);
    system (cmd);
    
    sprintf (cmd, "rm %s", fViffG);
    system (cmd);

    sprintf (fAscB, "mdTMP%dB.asc", i); 
    sprintf (cmd, "vprdata -i %s -f %s", fViffB, fAscB);
    system (cmd);
    
    sprintf (cmd, "rm %s", fViffB);
    system (cmd);

    /* === create feature file === */
    /* === and get rid of tmp files === */

    printf ("writing to feature file: %s...\n", FEAT_FILE);
    sprintf (fAsc, "mdTMP%d", i); 
    Write2FeatFile(fAsc, tmpClass, PIX_PCT, FEAT_FILE);

    sprintf (cmd, "rm %s", fAscR); system (cmd);
    sprintf (cmd, "rm %s", fAscG); system (cmd);
    sprintf (cmd, "rm %s", fAscB); system (cmd);

    i++; 

    /* === get filenames for next sequence === */

    sprintf (fViff0, "mdTMP%d.viff0", i); 
    sprintf (fViff1, "mdTMP%d.viff1", i); 
    sprintf (fViffR, "mdTMP%dR.viff", i);   
    sprintf (fViffG, "mdTMP%dG.viff", i); 
    sprintf (fViffB, "mdTMP%dB.viff", i); 

  }
  
  
}


main (int argc, char *argv[])

{
  /* === show usage === */
  if (argc != 3) {
    fprintf(stdout, "\nusage: mdtrain IMAGE FeatureFile\n");
    return (-1);
  }

  IN_IMG = argv[1];
  FEAT_FILE = argv[2];

  /* === convert input image to tmp.viff === */
  printf ("converting image to viff...\n"); 
  ConvertInImg2Viff();

  /* === run editimage for extracting training regions === */
  printf ("invoking image editor...\n");
  printf ("from IMAGE UTILITIES, pull down REGION-OF-INTEREST...\n");
  printf ("in `(write) Filename', replace `#' by sub-img# (IN SEQUENCE) & `*' by class\n");
  system ("editimage -i tmp.viff -roi_out mdTMP#.viff*");

  /* === get rid of tmp.viff === */
  printf ("getting rid of tmp files...\n");
  system ("rm tmp.viff");

  /* === create ascii RGB files === */
  CreateFeatFromImg();

  printf ("done...\n");
}


